from astropy.io import fits


def createFitFile(fitsHeader, fitsData, fileName):
    fits.writeto(fileName, fitsData, header=fitsHeader, overwrite=True)


def createFitFileWithSameHeader(fitsData, fileName, fileNameToCopyHeaderFrom):
    with fits.open(fileNameToCopyHeaderFrom) as fd:
        headerToCopy = fd[0].header
        createFitFile(headerToCopy, fitsData, fileName)



import datetime
import re
from datetime import date
from pathlib import Path
from typing import Dict, Iterable, Tuple

import numpy as np

from m23.constants import OBSERVATION_DATETIME_FORMAT, SKY_BG_FILENAME_DATE_FORMAT
from m23.sky import angle_of_elevation
from m23.sky.moon import moon_distance
from m23.sky.moon import phase as get_moon_phase_name
from m23.sky.moon import position as get_moon_phase
from m23.utils import half_round_up_to_int


class SkyBgFile:
    file_name_re = re.compile(r"(\d{2}-\d{2}-\d{2})_m23_(\d+\.\d*)_sky_bg\.txt")
    date_observed_datetime_format = OBSERVATION_DATETIME_FORMAT

    # We divide up the sky into several square sections (currently, sized 64px)
    # and calculate the mean background in that region.  Additionally,
    # background data is associated with date and time at which it is
    # calculated.
    DateTimeType = str
    BackgroundRegionType = Tuple[int, int]
    BGAduPerPixelType = float
    BackgroundDictType = Dict[BackgroundRegionType, BGAduPerPixelType]
    SkyBGDataType = Iterable[Tuple[DateTimeType, BackgroundDictType]]

    @classmethod
    def generate_file_name(cls, night_date: date):
        """
        Returns the file name for the sky background for a night
        param : night_date: Date for the night
        """
        return f"{night_date.strftime(SKY_BG_FILENAME_DATE_FORMAT)}_m23_sky_bg.txt"

    def __init__(self, path: str | Path) -> None:
        if type(path) == str:
            path = Path(path)
        self.__path = path
        self.__data = None

    def path(self):
        return self.__path

    def _validate_file(self):
        if not self.path().exists():
            raise FileNotFoundError(f"File not found {self.path()}")
        if not self.path().is_file():
            raise ValueError("Directory provided, expected file f{self.path()}")

    def create_file(
        self,
        sky_bg_data: SkyBGDataType,
        color_normfactors_title: Iterable[str],
        color_normfactors_values: Iterable[float],
        brightness_normfactors_title: Iterable[str],
        brightness_normfactors_values: Iterable[float],
        first_img_number: int,
        last_img_number: int,
        normalized_cluster_angle: int,
    ):
        """
        Creates/Updates sky background file based on the `sky_bg_data`
        """
        if len(sky_bg_data) == 0:
            # Nothing to do
            with open(self.path(), "w") as fd:
                pass
            return

        color_normfactors_title_str = " ".join(map("{:<30s}".format, color_normfactors_title))
        brightness_normfactors_title_str = " ".join(
            map("{:<30s}".format, brightness_normfactors_title)
        )
        color_normfactors_values_str = " ".join(map("{:<30.5f}".format, color_normfactors_values))
        brightness_normfactors_values_str = " ".join(
            map("{:<30.5f}".format, brightness_normfactors_values)
        )

        bg_sections = map(lambda x: "_".join(map(str, x)), sky_bg_data[0][1].keys())
        bg_sections_str = "".join(map("{:<10s}".format, bg_sections))

        with open(self.path(), "w") as fd:
            fd.write(
                f"{'Date':<26s}"
                f"{'Image_number':<15s}"
                f"{'First_img':<15s}"  # Img number of the first logfile combined used
                f"{'Last_img':<15s}"  # Img number of the last logfile combined used
                f"{'Moon_Phase':<16s}"
                f"{'Moon_Phase_Name':<20s}"
                f"{'Normalized_cluster_angle':<30s}"
                f"{'Cluster_Angle_Round':<20s}"
                f"{'Cluster_Angle':<20s}"
                f"{'Cluster_Angle_Uncertainty':<30s}"
                f"{'Moon_Distance':<20s}"
                f"{'Mean':<10s}{'Median':<10s}"
                f"{'Std':<10s}"
                f"{color_normfactors_title_str}"
                f"{brightness_normfactors_title_str}"
                f"{bg_sections_str}\n"
            )
            for night_datetime, bg_data, img_number in sky_bg_data:
                bg_data_np = np.array([bg_data[x] for x in bg_data.keys()])

                # We ignore the bogus values before taking the mean and the median
                bg_data_ignoring_bogus_values = bg_data_np[(bg_data_np > 0)]

                mean, median = np.mean(bg_data_ignoring_bogus_values), np.median(
                    bg_data_ignoring_bogus_values
                )
                std = np.std(bg_data_ignoring_bogus_values)
                date_time_of_observation = datetime.datetime.strptime(
                    night_datetime, self.date_observed_datetime_format
                )
                moon_phase = get_moon_phase(date_time_of_observation)
                moon_phase_name = get_moon_phase_name(date_time_of_observation)

                values_str = "".join(map("{:<10.2f}".format, bg_data_np))
                cluster_angle, cluster_angle_uncertainty = angle_of_elevation(
                    date_time_of_observation
                )  # noqa
                moon_dist_degrees = moon_distance(date_time_of_observation)
                fd.write(
                    f"{night_datetime:<26s}"
                    f"{img_number:<15d}"
                    f"{first_img_number:<15d}"
                    f"{last_img_number:<15d}"
                    f"{moon_phase:<16.5f}"
                    f"{moon_phase_name:<20s}"
                    f"{normalized_cluster_angle:<30d}"
                    f"{half_round_up_to_int(cluster_angle):<20.0f}"
                    f"{cluster_angle:<20.1f}"
                    f"{cluster_angle_uncertainty:<30.1f}"
                    f"{moon_dist_degrees:<20.2f}"
                    f"{mean:<10.2f}{median:<10.2f}"
                    f"{std:<10.2f}"
                    f"{color_normfactors_values_str}"
                    f"{brightness_normfactors_values_str}"
                    f"{values_str}\n"
                )

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"SkyBackgroundFile {self.path()}"



def blockRegions(arr, blockSize):
    """
    Taken from
    https://stackoverflow.com/questions/44782476/split-a-numpy-array-both-horizontally-and-vertically

    @param arr: numpy array to turn into blocks
    @param blockSize: tuple/array of rows/cols in a block
    """
    m, n = arr.shape
    M, N = blockSize
    return arr.reshape(m // M, M, n // N, N).swapaxes(1, 2).reshape(-1, M, N)



import logging
import sys
from pathlib import Path

import multiprocess as mp

from m23 import __version__
from m23.constants import FLUX_LOGS_COMBINED_FOLDER_NAME
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.file.reference_log_file import ReferenceLogFile
from m23.processor.process_nights import normalization_helper
from m23.utils import get_date_from_input_night_folder_name, get_log_file_name

from .renormalize_config_loader import (
    RenormalizeConfig,
    validate_renormalize_config_file,
)


def renormalize_auxiliary(renormalize_dict: RenormalizeConfig):
    def night_renorm_mapper(night):
        NIGHT_FOLDER = night["path"]
        night_date = get_date_from_input_night_folder_name(NIGHT_FOLDER.name)
        log_file_path = NIGHT_FOLDER / get_log_file_name(night_date)
        radii_of_extraction = renormalize_dict["processing"]["radii_of_extraction"]

        logger = logging.getLogger("LOGGER_" + str(night_date))
        logger.setLevel(logging.INFO)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        ch = logging.FileHandler(log_file_path)
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        # Write to std out in addition to writing to a logfile
        ch2 = logging.StreamHandler(sys.stdout)
        ch2.setFormatter(formatter)
        logger.addHandler(ch2)  # Write to stdout

        first_image = night["first_logfile_number"]
        last_image = night["last_logfile_number"]
        logger.info(
            f"Running renormalization for radii {radii_of_extraction}. Img: {first_image}-{last_image} with m23 version: {__version__}"  # noqa ES501
        )

        FLUX_LOGS_COMBINED_FOLDER: Path = NIGHT_FOLDER / FLUX_LOGS_COMBINED_FOLDER_NAME
        # Create log files combined folder if it doesn't yet exist
        FLUX_LOGS_COMBINED_FOLDER.mkdir(exist_ok=True)

        reference_log_file = ReferenceLogFile(
            renormalize_dict["reference"]["file"],
        )
        logfile_combined_reference_logfile = LogFileCombinedFile(
            renormalize_dict["reference"]["logfile"]
        )

        # Ensure color ref file path is str
        color_ref_file_path = str(renormalize_dict["reference"]["color"])

        log_files_to_use = [LogFileCombinedFile(file) for file in night["files_to_use"]]
        img_duration = log_files_to_use[0].img_duration()

        # Perform intranight normalization then internight normalization
        normalization_helper(
            radii_of_extraction,
            reference_log_file,
            log_files_to_use,
            img_duration,
            night_date,
            color_ref_file_path,
            NIGHT_FOLDER,
            logfile_combined_reference_logfile,
        )

    with mp.Pool() as p:
        p.map(night_renorm_mapper, renormalize_dict["input"]["nights"])


def renormalize(file_path: str):
    """
    Starts renormalization with the configuration file `file_path` provided as the argument.
    Calls auxiliary function `renormalize_auxiliary` if the configuration is valid.
    """
    validate_renormalize_config_file(Path(file_path), on_success=renormalize_auxiliary)



from datetime import date
from pathlib import Path

from m23.align import AlignmentTransformationType
from m23.constants import ALIGNED_STATS_FILE_DATE_FORMAT


class AlignmentStatsFile:
    @classmethod
    def generate_file_name(cls, night_date: date) -> str:
        """
        Generates filename based on the given `night_date`
        """
        return f"m23_aligned_stats_{night_date.strftime(ALIGNED_STATS_FILE_DATE_FORMAT)}.txt"

    def __init__(self, file_path) -> None:
        self.__path = Path(file_path)
        self.__is_read = False

    def path(self):
        return self.__path

    def create_file_and_write_header(self):
        """
        Create a file (wipes out if the file already exists)
        and writes header in the file
        """
        with open(self.path(), "w") as fd:
            fd.write(
                f"{'Image_Name':<30s}"
                f"{'Rotation':<20s}"
                f"{'Translation_X':<15s}"
                f"{'Translation_Y':<15s}"
                f"{'Scale':<10s}\n"
            )

    def add_record(self, image_name: str, alignment_stats: AlignmentTransformationType):
        """
        Adds a record to the file based on the provided `alignment_stats`
        """
        rotation, translation_x, translation_y, scale = alignment_stats
        with open(self.path(), "a") as fd:
            fd.write(
                f"{image_name:<30}"
                f"{rotation:<20.9f}"
                f"{translation_x:<15.3f}"
                f"{translation_y:<15.3f}"
                f"{scale:<10.3f}\n"
            )

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"Aligned combined file: {self.path()}"



import datetime
import os
import sys
from datetime import date
from pathlib import Path
from typing import Callable, Dict, List, TypedDict

import toml
from typing_extensions import NotRequired

from m23.constants import (
    CAMERA_CHANGE_2022_DATE,
    INPUT_CALIBRATION_FOLDER_NAME,
    M23_RAW_IMAGES_FOLDER_NAME,
    TYPICAL_NEW_CAMERA_CROP_REGION,
)
from m23.exceptions import InvalidDatetimeInConfig
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.reference import get_reference_files_dict
from m23.utils import (
    get_darks,
    get_date_from_input_night_folder_name,
    get_flats,
    get_raw_images,
)


# TYPE related to Config object described by the configuration file
class ConfigImage(TypedDict):
    rows: int
    columns: int
    crop_region: NotRequired[List[List[List[int]]]]


class ConfigProcessing(TypedDict):
    no_of_images_to_combine: int
    radii_of_extraction: List[int]


class ConfigInputNight(TypedDict):
    path: str | Path
    masterflat: NotRequired[str]
    starttime: NotRequired[str]


class ConfigInput(TypedDict):
    nights: List[ConfigInputNight]
    radii_of_extraction: List[int]


class ConfigReference(TypedDict):
    image: str | Path
    file: str | Path
    logfile: str | Path
    color: str | Path


class ConfigOutput(TypedDict):
    path: str | Path
    save_aligned: NotRequired[bool]
    save_calibrated: NotRequired[bool]


class ConfigDateTime(TypedDict):
    start: datetime.datetime
    end: datetime.datetime


class Config(TypedDict):
    image: ConfigImage
    processing: ConfigProcessing
    reference: ConfigReference
    input: ConfigInput
    output: ConfigOutput


def is_valid_radii_of_extraction(lst):
    """Verifies that each radius of extraction is a positive integer"""
    is_valid = all([type(i) == int and i > 0 for i in lst])
    if not is_valid:
        sys.stderr.write.write("Radius of extraction must be positive integers\n")
    return is_valid


def create_processing_config(config_dict: Config) -> Config:  # noqa
    """
    Mutates `config_dict` to :
    1. Provide default values if optional values aren't provided.
    2. Replace all path str with Path objects
    """
    # Add empty list as the crop region if not present
    if not config_dict["image"].get("crop_region"):
        config_dict["image"]["crop_region"] = []

    # Convert input night str to Path objects
    for night in config_dict["input"]["nights"]:
        if type(night["path"]) == str:
            night["path"] = Path(night["path"])
        if night.get("masterflat") and type(night["masterflat"] == str):
            night["masterflat"] = Path(night["masterflat"])

    # Convert output path to Path object
    if type(config_dict["output"]["path"]) == str:
        config_dict["output"]["path"] = Path(config_dict["output"]["path"])

    # Add a boolean specifying whether or not to save aligned and calibrated images.
    if config_dict["output"].get("save_aligned"):
        config_dict["output"]["save_aligned"] = True
    else:
        config_dict["output"]["save_aligned"] = False

    if config_dict["output"].get("save_calibrated"):
        config_dict["output"]["save_calibrated"] = True
    else:
        config_dict["output"]["save_calibrated"] = False

    # Convert reference file/img to Path object
    if type(config_dict["reference"]["file"]) == str:
        config_dict["reference"]["file"] = Path(config_dict["reference"]["file"])
    if type(config_dict["reference"]["image"]) == str:
        config_dict["reference"]["image"] = Path(config_dict["reference"]["image"])
    if type(config_dict["reference"]["color"]) == str:
        config_dict["reference"]["color"] = Path(config_dict["reference"]["color"])

    # Remove duplicates radii of extraction
    radii = list(set(config_dict["processing"]["radii_of_extraction"]))
    config_dict["processing"]["radii_of_extraction"] = radii

    return config_dict


def prompt_to_continue(msg: str):
    sys.stderr.write(msg + "\n")
    response = input("Do you want to continue (y/yes to continue): ")
    if response.upper() not in ["Y", "YES"]:
        os._exit(1)


def sanity_check_image(config: ConfigImage, night_date: date):  # noqa
    """
    Checks for abnormal values in configuration images
    """
    rows, cols = config["rows"], config["columns"]
    crop_region = config.get("crop_region")
    old_camera = night_date < CAMERA_CHANGE_2022_DATE
    if old_camera:
        if rows != 1024:
            prompt_to_continue("Detected non 1024 image row value for old camera date")
        if cols != 1024:
            prompt_to_continue("Detected non 1024 image column value for old camera date")
        if crop_region and type(crop_region) == list and len(crop_region) > 0:
            prompt_to_continue("Detected use of crop region for old camera.")
    else:
        if rows != 2048:
            prompt_to_continue("Detected non 2048 image row value for new camera date")
        if cols != 2048:
            prompt_to_continue("Detected non 2048 image column value for new camera date")
        if (
            not crop_region
            or crop_region
            and type(crop_region) != list
            or type(crop_region) == list
            and len(crop_region) == 0
        ):
            prompt_to_continue(
                "We typically use crop images from new camera, you don't seem to define it"
            )
        else:
            try:
                for crop_section_index, crop_section in enumerate(crop_region):
                    for (
                        section_coordinate_index,
                        section_coordinate,
                    ) in enumerate(crop_section):
                        if (
                            section_coordinate
                            != TYPICAL_NEW_CAMERA_CROP_REGION[crop_section_index][
                                section_coordinate_index
                            ]
                        ):
                            prompt_to_continue(
                                "Mismatch between default crop region"
                                + " used in new camera and config file."
                            )
                            # Ignore further checking if already made the user
                            # aware of inconsistency once
                            return

            except Exception as e:
                prompt_to_continue(
                    f"Error while checking crop region with standard crop region value. {e}"
                )


def sanity_check(config_dict: Config) -> Config:
    """
    This method performs any sanity checks on the configuration file.
    """
    # Ensure sane values for rows/cols, etc.
    for night in config_dict["input"]["nights"]:
        night_date = get_date_from_input_night_folder_name(night["path"])
        sanity_check_image(config_dict["image"], night_date)
    return config_dict


def verify_optional_image_options(options: Dict) -> bool:
    """
    Verifies that the optional image options are valid
    """
    if len(options.keys()) > 1:
        return False
    crop_region: List[List[int]] = options.get("crop_region", [])
    # Ensure that all values in crop_region are non-negative integers
    try:
        for i in crop_region:
            for j in i:
                valid_values = all([type(x) == int and x >= 0 for x in j])
                if not valid_values:
                    sys.stderr.write(f"Invalid value detected in crop_region {j}.\n")
                    return False
    except ValueError:
        sys.stderr.write(f"Error in crop_region {j}.\n")
        return False
    return True  # Valid


def is_night_name_valid(NIGHT_INPUT_PATH: Path):
    """
    Returns if the input night folder name follows naming conventions.
    Prints msg to stderr if invalid.
    """
    # Check if the name of input folder matches the convention
    try:
        get_date_from_input_night_folder_name(NIGHT_INPUT_PATH.name)
        return True
    except Exception:  # noqa
        sys.stderr.write(
            f"Night {NIGHT_INPUT_PATH} folder name doesn't match the naming convention\n"
        )
        return False


def validate_night(night: ConfigInputNight) -> bool:  # noqa
    """
    Checks whether the input configuration provided for night is valid.
    We check whether the input folders follow the required conventions,
    whether the right files are present and more.
    """
    try:
        NIGHT_INPUT_PATH = Path(night["path"])
    except Exception:  # noqa
        sys.stderr.write(f"Invalid night {night} in config file.\nCheck path spell\n")
        return False

    # Check if the night input path exists
    if not NIGHT_INPUT_PATH.exists():
        sys.stderr.write(f"Images path for {night} doesn't exist\n")
        return False

    if not is_night_name_valid(NIGHT_INPUT_PATH):
        return False

    CALIBRATION_FOLDER_PATH = NIGHT_INPUT_PATH / INPUT_CALIBRATION_FOLDER_NAME
    # Check if Calibration Frames exists
    if not CALIBRATION_FOLDER_PATH.exists():
        sys.stderr.write(f"Path {CALIBRATION_FOLDER_PATH} doesn't exist\n")
        return False

    M23_FOLDER_PATH = NIGHT_INPUT_PATH / M23_RAW_IMAGES_FOLDER_NAME
    # Check if m23 folder exists
    if not M23_FOLDER_PATH.exists():
        sys.stderr.write(f"Path {M23_FOLDER_PATH} doesn't exist\n")
        return False

    # Check for flats
    # Either the masterflat should be provided or the night should contain its own flats.
    if night.get("masterflat"):
        if not Path(night["masterflat"]).exists():
            sys.stderr.write(f"Provided masterflat path for {night} doesn't exist.\n")
            return False
    # If masterflat isn't provided, the night should have flats to use
    elif len(list(get_flats(CALIBRATION_FOLDER_PATH))) == 0:
        sys.stderr.write(
            f"Night {night} doesn't contain flats in {CALIBRATION_FOLDER_PATH}."
            + " Provide masterflat path.\n"
        )
        return False

    # Check for darks
    if len(list(get_darks(CALIBRATION_FOLDER_PATH))) == 0:
        sys.stderr.write(
            f"Night {night} doesn't contain darks in {CALIBRATION_FOLDER_PATH}."
            + " Cannot continue without darks.\n"
        )
        return False

    # Check for raw images
    try:
        if len(list(get_raw_images(M23_FOLDER_PATH))) == 0:
            sys.stderr.write(f"Night {night} doesn't have raw images in {M23_FOLDER_PATH}.\n")
            return False
    except ValueError as e:
        sys.stderr.write(
            "Raw image in night {night} doesn't confirm to 'something-00x.fit' convention.\n"
        )
        raise e

    # Validate the start and end time of observation if provided
    # Check if datetime is declared
    if start := night.get("starttime"):
        try:
            validate_datetime(start)
        except InvalidDatetimeInConfig as e:
            sys.stderr.write(
                f"OPTIONAL observation start time for {night} isn't"
                " in the format  1979-05-27T07:32:00 where timezone is UT\n"
            )
            raise e

    return True  # Assuming we did the best we could to catch errors


def validate_input_nights(list_of_nights: List[ConfigInputNight]) -> bool:
    """
    Returns True if input for all nights is valid, False otherwise.
    """
    return all([validate_night(night) for night in list_of_nights])


def validate_reference_files(
    reference_image: str,
    reference_file: str,
    color_ref_file: str,
    logfile: str,
    radii: List[int],
) -> bool:
    """
    Returns True if reference_image and reference_file paths exist
    """
    img_path = Path(reference_image)
    file_path = Path(reference_file)
    color_path = Path(color_ref_file)
    logfile_path = Path(logfile)
    if not (img_path.exists() and img_path.is_file() and img_path.suffix == ".fit"):
        sys.stderr.write("Make sure that the reference image exists and has .fit extension\n")
        return False
    if not (file_path.exists() and file_path.is_file() and file_path.suffix == ".txt"):
        sys.stderr.write("Make sure that the reference file exists and has .txt extension\n")
        return False
    if not (color_path.exists() and color_path.is_file() and color_path.suffix == ".txt"):
        sys.stderr.write("Make sure that the color reference file exists and has .txt extension\n")
        return False
    if not (logfile_path.exists() and logfile_path.is_file() and logfile_path.suffix == ".txt"):
        sys.stderr.write("Make sure that the log file exists and has .txt extension\n")
        return False

    # Make sure that the logfile combined reference file has
    # all radii of extraction data
    available_radii = LogFileCombinedFile(logfile_path).get_star_data(1).radii_adu.keys()
    for i in radii:
        if i not in available_radii:
            sys.stderr.write(
                f"Radius {i} ADU data not present in provided logfile combined file. \n"
            )
            return False

    return True


def validate_datetime(time_obj):
    if not isinstance(time_obj, datetime.datetime):
        raise InvalidDatetimeInConfig


def verify_optional_output_options(output_options: Dict[str, any]):
    valid_keys = ["save_aligned", "save_calibrated"]
    for key in output_options.keys():
        if key not in valid_keys:
            sys.stderr.write(
                f"""Found invalid key {key} in output options configuration.
                Valid keys are: {valid_keys}"""
            )
            return False
    if save_aligned := output_options.get("save_aligned"):
        if not isinstance(save_aligned, bool):
            sys.stderr.write(
                f"Expected (true/false) instance for save_aligned option found {save_aligned}\n"
            )
            return False
    if save_calibrated := output_options.get("save_calibrated"):
        if not isinstance(save_aligned, bool):
            sys.stderr.write(
                "Expected (true/false) instance for save_calibrated"
                f" option found {save_calibrated}\n"
            )
            return False
    return True


def validate_file(file_path: Path, on_success: Callable[[Config], None]) -> None:
    """
    This method reads data processing configuration from the file path
    provided and calls the unary function on_success if the configuration
    file is valid with the configuration dictionary (Note, *not* config file).
    """
    if not file_path.exists() or not file_path.exists():
        raise FileNotFoundError("Cannot find configuration file")
    configuration = toml.load(file_path)
    load_configuration_with_necessary_reference_files(configuration)
    match configuration:
        case {
            "image": {
                "rows": int(_),
                "columns": int(_),
                **optional_image_options,
            },
            "processing": {
                "no_of_images_to_combine": int(_),
                "radii_of_extraction": list(radii_of_extraction),
            },
            "reference": {
                "image": str(reference_image),
                "file": str(reference_file),
                "logfile": str(logfile),
                "color": str(color_ref_file),
            },
            "input": {"nights": list(list_of_nights)},
            "output": {"path": str(_), **optional_output_options},
        } if (
            verify_optional_image_options(optional_image_options)
            and verify_optional_output_options(optional_output_options)
            and is_valid_radii_of_extraction(radii_of_extraction)
            and validate_input_nights(list_of_nights)
            and validate_reference_files(
                reference_image,
                reference_file,
                color_ref_file,
                logfile,
                radii_of_extraction,
            )
        ):
            # Check for the optional configurations
            # Optional configs should either not be declared or be
            # correctly declared
            on_success(sanity_check(create_processing_config(configuration)))
        case _:
            sys.stderr.write("Invalid format.\n")


def load_configuration_with_necessary_reference_files(configuration, pop=None):
    # Users need not define reference files optional
    reference_files_dict = get_reference_files_dict()
    if user_defined_reference := configuration.get("reference"):
        # If user defines any reference file, that takes precedence
        # over files configured to be used as default
        for key in reference_files_dict:
            if key not in user_defined_reference:
                user_defined_reference[key] = reference_files_dict[key]
    else:
        configuration["reference"] = reference_files_dict
    if pop:
        for item_to_pop in pop:
            configuration["reference"].pop(item_to_pop)



import re
from collections import namedtuple
from datetime import date, datetime
from pathlib import Path
from typing import Dict, Iterable, Tuple

import numpy as np
import numpy.typing as npt

from m23.constants import OBSERVATION_DATETIME_FORMAT
from m23.file.aligned_combined_file import AlignedCombinedFile
from m23.sky import angle_of_elevation


# Note that LogFileCombined is the one that that has the data for aligned combined
# images after extracting stars from the image. This is not to be confused with FluxLogCombined
# that is created after intra-night normalization (note *not* internight normalization)
class LogFileCombinedFile:
    # Class attributes
    header_rows = 9
    data_titles_row_zero_index = 8
    date_format = "%m-%d-%y"
    sky_adu_column = 5
    x_column = 0
    y_column = 1
    file_name_re = re.compile(r"(\d{2}-\d{2}-\d{2})_m23_(\d+\.\d*)-(\d{3})\.txt")
    star_adu_radius_re = re.compile(r"Star ADU (\d+)")

    StarLogfileCombinedData = namedtuple(
        "StarLogfileCombinedData",
        ["x", "y", "xFWHM", "yFWHM", "avgFWHM", "sky_adu", "radii_adu"],
    )
    LogFileCombinedDataType = Dict[int, StarLogfileCombinedData]

    date_observed_datetime_format = OBSERVATION_DATETIME_FORMAT

    @classmethod
    def generate_file_name(cls, night_date: date, img_no: int, img_duration: float):
        """
        Returns the file name to use for a given star night for the given night date
        param : night_date: Date for the night
        param: img_no : Image number corresponding to the Aligned combined image
        param: img_duration : the duration of images taken on the night
        """
        return f"{night_date.strftime(cls.date_format)}_m23_{img_duration}-{img_no:03}.txt"

    def __init__(self, file_path: str) -> None:
        self.__path = Path(file_path)
        self.__is_read = False
        self.__header = None
        self.__data = None
        self.__title_row = None
        self.__cluster_angle = None

    def _read(self):
        with self.__path.open() as fd:
            lines = [line.strip() for line in fd.readlines()]
            # Save the title row
            # We split the title row by gap of more than two spaces
            self.__title_row = re.split(r"\s{2,}", lines[self.data_titles_row_zero_index])
            self.__header = lines[: self.header_rows]
            lines = lines[self.header_rows :]  # Skip headers - 1
            # Create a 2d list
            lines = [line.split() for line in lines]
            # Convert to 2d numpy array
            self.__data = np.array(lines, dtype="float")
        self.__is_read = True

    def _title_row(self):
        if not self.__is_read:
            self._read()
        return self.__title_row

    def _adu_radius_header_name(self, radius: int):
        return f"Star ADU {radius}"

    def _get_column_number_for_adu_radius(self, radius: int):
        titles = self._title_row()
        return titles.index(self._adu_radius_header_name(radius))

    def get_cluster_angle_with_uncertainty(self) -> Tuple[float, float]:
        """
        Return the cluster angle along with uncertainty of measurement of the
        time the image was taken
        """
        if not self.__cluster_angle:
            self.__cluster_angle = angle_of_elevation(
                datetime.strptime(self.datetime(), OBSERVATION_DATETIME_FORMAT)
            )
        # Note that cluster angle contains the angle as well as uncertainty
        return self.__cluster_angle

    def get_cluster_angle(self) -> float:
        """Return the cluster angle of the time the image was taken"""
        return self.get_cluster_angle_with_uncertainty()[0]

    def is_valid_file_name(self) -> bool:
        """
        Checks if the filename matches the naming convention
        """
        return bool(self.file_name_re.match(self.path().name))

    def header(self) -> Iterable[str]:
        """
        Returns an iterable to string representing the header information in the
        file
        """
        if not self.__is_read:
            self._read()
        return self.__header

    def datetime(self) -> str:
        """
        Return the datetime string representing the observation of this image
        or empty string if no data is present
        """
        return self.header()[0].split("\t")[-1]

    def night_date(self) -> date | None:
        """
        Returns the night date that can be inferred from the file name
        """
        if self.is_valid_file_name():
            # The first capture group contains the night date
            return datetime.strptime(
                self.file_name_re.match(self.path().name)[1],
                self.date_format,
            ).date()

    def get_adu(self, radius: int):
        """
        Returns an ordered array of ADU for stars for given `radius`.
        The first row of the array is the adu of star 1, 200th row for star 200,
        and the like
        """
        radius_col = self._get_column_number_for_adu_radius(radius)
        return self.data()[:, radius_col]

    def get_sky_adu_column(self):
        """
        Returns an ordered array of stars sky adu.
        The first row of the array is the sky adu of star 1, 200th row for star 200,
        and the like
        """
        return self.data()[:, self.sky_adu_column]

    def get_x_position_column(self):
        """
        Returns an ordered array of stars x positions.
        The first row of the array is the x position of star 1, 200th row for star 200,
        and the like
        """
        return self.data()[:, self.x_column]

    def get_y_position_column(self) -> npt.NDArray:
        """
        Returns an ordered array of stars y positions.
        The first row of the array is the y position of star 1, 200th row for star 200,
        and the like
        """
        return self.data()[:, self.y_column]

    def get_star_data(self, star_no: int) -> StarLogfileCombinedData:
        """
        Returns the details related to a particular `star_no`
        Returns a named tuple `StarLogfileCombinedData`
        """
        star_data = self.data()[star_no - 1]
        titles = self._title_row()
        first_radii_adu_column = 6
        radii_adu = {}
        for index, col_name in enumerate(titles[first_radii_adu_column:]):
            radius = int(self.star_adu_radius_re.match(col_name)[1])
            radii_adu[radius] = star_data[first_radii_adu_column + index]
        return self.StarLogfileCombinedData(*star_data[:first_radii_adu_column], radii_adu)

    def img_duration(self) -> float | None:
        """
        Returns the image duration that can be inferred from the file name
        """
        if self.is_valid_file_name():
            # The second capture group contains the image duration
            return float(self.file_name_re.match(self.path().name)[2])

    def img_number(self) -> int | None:
        """
        Returns the image number associated to the filename if the file name is valid
        """
        if self.is_valid_file_name():
            # The third capture group contains the image number
            return int(self.file_name_re.match(self.path().name)[3])

    def is_file_format_valid(self):
        """
        Checks if the file format is valid
        """
        return True

    def path(self):
        return self.__path

    def data(self):
        if not self.__is_read:
            self._read()
        # Return the copy of the data so that the a caller
        # doesn't get affected by another misbehaving caller.
        return np.copy(self.__data)

    def create_file(
        self,
        data: LogFileCombinedDataType,
        aligned_combined_file: AlignedCombinedFile,
        datetime_of_img: str = "",
    ):
        """
        Creates logfile combined file based on the provided data

        param: data : Data that's to be written
        param: aligned_combined_file: AlignedCombinedFile on which this file is based
        """
        if not self.is_valid_file_name():
            raise ValueError(f"Invalid file name {self.path()}")

        if len(data) == 0:
            radii = []
        else:
            radii = data[list(data.keys())[0]].radii_adu.keys()

        stars = sorted(data.keys())
        no_of_stars = len(stars)
        with self.path().open("w") as fd:
            # First line represents the datetime
            fd.write(f"ObservedAt:\t{datetime_of_img}\n")
            fd.write("Star Data Extractor Tool: (Note: This program mocks format of AIP_4_WIN) \n")
            fd.write(f"\tImage {aligned_combined_file.path().name}\n")
            fd.write(f"\tTotal no of stars: {no_of_stars}\n")
            fd.write(f"\tRadius of star diaphragm: {', '.join(map(str, radii))}\n")
            fd.write("\tSky annulus inner radius: \n")
            fd.write("\tSky annulus outer radius: \n")
            fd.write("\tThreshold factor: \n")

            headers = [
                "X",
                "Y",
                "XFWHM",
                "YFWHM",
                "Avg FWHM",
                "Sky ADU",
            ] + [self._adu_radius_header_name(radius) for radius in radii]
            for header in headers:
                fd.write(f"{header:>16s}")
            fd.write("\n")

            for star in stars:  # Sorted in ascending order by star number
                star_data = data[star]
                fd.write(
                    f"{star_data.x:>16.2f}{star_data.y:>16.2f}{star_data.xFWHM:>16.4f}{star_data.yFWHM:>16.4f}{star_data.avgFWHM:>16.4f}{star_data.sky_adu:>16.2f}"  # noqa
                )
                for radius in radii:
                    fd.write(f"{star_data.radii_adu[radius]:16.2f}")
                fd.write("\n")

    def __len__(self):
        """Returns the number of stars present in the dataset"""
        return len(self.data())

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"Log file combined: {self.__path}"



import re
from datetime import date, datetime
from pathlib import Path
from typing import Iterable

import numpy as np
import numpy.typing as npt

from m23.constants import FLUX_LOG_COMBINED_FILENAME_DATE_FORMAT
from m23.file.normfactor_file import NormfactorFile
from m23.file.reference_log_file import ReferenceLogFile


# Note that FluxLogCombined is the one that we have for multiple
# radius of extraction. This is generated after intra-night (*not* inter-night)
# normalization. Don't confuse this file with LogFileCombined that
# is generated after extraction step.
class FluxLogCombinedFile:
    """
    This class is instantiated with the string representing
    file path for the Flux Log Combined file that you want to analyze

    This object could be useful for analyzing values like the attendance of a
    star on a particular night, it's mean and median values on the night, etc.
    """

    # Class attributes
    header_rows = 6  # Specifies the first x rows that don't contain header information
    file_name_re = re.compile("(\d{2}-\d{2}-\d{2})_m23_(\d+\.\d*)-(\d{1,4})_flux\.txt")  # noqa
    header_columns = ["ADU", "X", "Y", "Normfactors", "DateTime"]

    def __init__(self, path: str | Path) -> None:
        if type(path) == str:
            path = Path(path)
        self.__path = path
        self.__data = None
        self.__valid_data = None
        self.__read_data = False
        self.__attendance = None

    @classmethod
    def generate_file_name(cls, night_date: date, star_no: int, img_duration: float):
        """
        Returns the file name to use for a given star night for the given night date
        param : night_date: Date for the night
        param: star_no : Star number
        param: img_duration : the duration of images taken on the night
        """
        return f"{night_date.strftime(FLUX_LOG_COMBINED_FILENAME_DATE_FORMAT)}_m23_{img_duration}-{star_no:04}_flux.txt"  # noqa

    def _validate_file(self):
        if not self.path().exists():
            raise FileNotFoundError(f"File not found {self.path()}")
        if not self.path().is_file():
            raise ValueError("Directory provided, expected file f{self.path()}")

    def _calculate_attendance(self) -> float:
        """
        Calculates and returns the attendance for the night based on `self.data`
        Note that attendance is a value between 0-1.

        Preconditions:
            The object should have valid `self.data`
        Assumptions:
            `self.data` contains all data point albeit empty for a start for the
            night
        """
        data_points = len(self.data())
        positive_value_data_points = len(self.valid_data())
        return positive_value_data_points / data_points

    def is_attendance_over_half(self) -> bool:
        """
        Returns whether the attedance is over half using the same
        method as the IDL code does. Note that this code isn't exactly
        correct because of the way it does rounding.
        """
        data_points = len(self.data())
        positive_value_data_points = len(self.valid_data())
        # Note the usage of integer division mimics IDL although mere float
        # division might be better.
        return positive_value_data_points >= (data_points // 2)

    # Accessors

    def read_file_data(self):
        """
        Reads the file and sets the the data attribute and attendance attribute
        in the object
        """
        self._validate_file()
        with self.path().open() as fd:
            lines = [line.strip() for line in fd.readlines()]
            lines = lines[self.header_rows :]  # Skip the header rows

            # Create a 2d list
            lines = [line.split() for line in lines]

            # Convert to 2d numpy array
            self.__data = np.array(lines)

            self.__all_adus = np.array(self.__data[:, 0], dtype="float")

            # These might be made public future but are unstable
            # # thus not available as API at the moment
            # self.__all_x_values = np.array(self.__data[:, 1], dtype="float")
            # self.__all_y_values = np.array(self.__data[:, 2], dtype="float")
            # self.__all_normfactors = np.array(self.__data[:, 3], dtype="float")
            # self.__all_dates = np.array(self.__data[:, 4])

            # Remove nan and values < 0
            self.__valid_adus = self.__all_adus[self.__all_adus > 0]

        self.__read_data = True  # Marks file as read
        self.__attendance = self._calculate_attendance()

    def is_valid_file_name(self) -> bool:
        """
        Checks if the file name is valid as per the file naming conventions
        of m23 data processing library.
        """
        return bool(self.file_name_re.match(self.path().name))

    def night_date(self) -> date | None:
        """
        Returns the night date that can be inferred from the file name
        """
        if self.is_valid_file_name():
            # The first capture group contains the night date
            return datetime.strptime(
                self.file_name_re.match(self.path().name)[1],
                FLUX_LOG_COMBINED_FILENAME_DATE_FORMAT,
            ).date()

    def img_duration(self) -> float | None:
        """
        Returns the image duration that can be inferred from the file name
        """
        if self.is_valid_file_name():
            # The second capture group contains the image duration
            return float(self.file_name_re.match(self.path().name)[2])

    def star_number(self) -> int | None:
        """
        Returns the star number associated to the filename if the file name is valid
        """
        if self.is_valid_file_name():
            # The third capture group contains the star number
            return int(self.file_name_re.match(self.path().name)[3])

    def is_file_format_valid(self):
        """
        Checks if the file format is valid
        """
        return True

    def path(self) -> Path:
        return self.__path

    def data(self) -> None | npt.ArrayLike:
        """
        The data property returns either None or a numpy one dimensional array
        of the adu values
        """
        if not self.__read_data:
            self.read_file_data()
        # Note there we're returning just the ADUs not the entire records
        return self.__all_adus

    def valid_data(self) -> None | npt.ArrayLike:
        """
        Returns a sample of adu data for the star for the nights with only valid
        data points > 0 magnitudes
        """
        if not self.__read_data:
            self.read_file_data()
        return self.__valid_adus

    def attendance(self) -> float:
        """
        Returns the attendance % (between 0-1) for star corresponding to the
        flux log combined file for the night
        """
        self._validate_file()
        if not self.__read_data:
            self.read_file_data()
        return self.__attendance

    def median(self) -> float:
        """
        Returns the median value for the star for the night
        Note that this is the median of only valid data points (> 0 magnitudes)
        This means that 0.00 values are automatically ignored.
        Note that if the night doesn't contain any valid data, this returns nan
        """
        self._validate_file()
        if not self.__read_data:
            self.read_file_data()
        return np.median(
            self.valid_data()
        )  # Note to use only the valid data points to calculate median

    def specialized_median_for_internight_normalization(self) -> float:
        """
        Returns specialized median flux for the night to be used by internight
        normalization code. This is special in that the median is only
        calculated from images that got applied internight normalization factor
        within a certain range. Additionally (as always) we ignore data points
        that are zero values when calculating median This is just an
        implementation of the way things are/were done in the IDL code.
        """
        min_tolerable_intranight_normfactor = 0.85
        max_tolerable_intranight_normfactor = 1.15

        # Get the *intra* night norm factors file from the same directory as this file is in
        parent_dir = self.path().parent
        normfactor_files = list(parent_dir.glob("*normfactor*"))
        if len(normfactor_files) == 0:
            raise ValueError("Normfactor file not found in {parent_dir}")
        if len(normfactor_files) > 1:
            raise ValueError("Multiple Normfactor files found in {parent_dir}")
        normfactor_file = NormfactorFile(normfactor_files[0].absolute())

        data_to_use = []
        for index, data in enumerate(self.data()):
            # Add the value only if it's > 0 and the normfactor for the image is
            # within specified range
            if (
                data > 0
                and min_tolerable_intranight_normfactor
                <= normfactor_file.data()[index]
                <= max_tolerable_intranight_normfactor
            ):
                data_to_use.append(data)

        # IDL style Median
        if len(data_to_use) == 0:
            return np.nan
        mid_value = len(data_to_use) // 2
        return sorted(data_to_use)[mid_value]

    def mean(self) -> float:
        """
        Returns the mean value for the star for the night
        Note that this is the mean of only valid data points (> 0 magnitudes)
        This means that 0.00 values are automatically ignored.
        Note that if the night doesn't contain any valid data, this returns nan
        """
        self._validate_file()
        if not self.__read_data:
            self.read_file_data()
        return np.mean(
            self.valid_data()
        )  # Note to use only the valid data points to calculate mean

    def create_file(
        self,
        adu_data: npt.NDArray,
        start_img: int,
        end_img: int,
        x_positions: Iterable[float],
        y_positions: Iterable[float],
        normfactors: Iterable[float],
        date_times: Iterable[str],
        reference_logfile: ReferenceLogFile,
    ):
        """
        Creates/Updates Flux Log Combined file

        param: data: The flux values for the star
        param: start_image: The first aligned combined image number used
        param: end_image: The last aligned combined image number used
        param: location: The x and y coordinates of the star in the image
            aligned combined image
        param: reference_logfile: The reference logfile used
        """
        if not self.is_valid_file_name():
            raise ValueError(f"File name is invalid {self.path()}")

        with self.path().open("w") as fd:
            fd.write("Program:\n")
            fd.write(f"Started with image\t{start_img}\n")
            fd.write(f"Ended with image\t{end_img}\n")
            fd.write(f"Reference log file used: {reference_logfile}\n")
            median_x, median_y = np.median(x_positions), np.median(y_positions)
            fd.write(f"MedianX,Y: \t{median_x:.3f}, {median_y:.3f}\n")
            fd.write(f"{'ADU':<15s}{'X':<13s}{'Y':<13s}{'Norm':<15s}{'Datetime':<32s}\n")

            cols_to_save = np.zeros(
                np.array(adu_data).size,
                dtype=[
                    ("adu", float),
                    ("x", float),
                    ("y", float),
                    ("norm", float),
                    ("date", "U32"),
                ],
            )
            cols_to_save["adu"] = adu_data
            cols_to_save["x"] = x_positions
            cols_to_save["y"] = y_positions
            cols_to_save["norm"] = normfactors
            cols_to_save["date"] = date_times

            np.savetxt(
                fd,
                cols_to_save,
                fmt="%-15.2f%-13.2f%-13.2f%-15.5f%-32s",
            )

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"FluxLogCombinedFile {self.path()}"



import cv2 as cv
import numpy as np


def fillMatrix(matrix, polygons, fillValue):
    """
    This is a wrapper our open cv's fillPoly function

    @param matrix: (2d) matrix to work with
    @param polygons: array of array of tuples (x, y) defining a polygon fill region
    @param fillValue: value to fill

    @return Mutated original `matrix`

    Example:
    matrix:  np.arange(50).reshape(2, 5, 5)[1]
    pol1: [(0,0), (1,1), (0,4)]
    pol2: [(4, 0), (3, 1), (4,4)]
    polygons: [pol1, pol2]
    fillValue: 0

    """
    return cv.fillPoly(matrix, np.array(polygons), fillValue)



from .generate_masterflat import generate_masterflat
from .nights_csv import create_nights_csv
from .process_nights import start_data_processing
from .renormalize import renormalize

__all__ = [
    "start_data_processing",
    "renormalize",
    "generate_masterflat",
    "create_nights_csv",
]



from datetime import datetime
from pathlib import Path
from typing import Dict, Tuple

import numpy as np
import numpy.typing as npt

from m23.file import line_str_contains_numbers_and_non_alphabets


class ReferenceLogFile:
    header_rows = 9  # The first x rows that don't contain data but metadata

    column_numbers: Dict[str, int] = {
        "x": 0,  # Beware of the zero indexing
        "y": 1,
        "sigma": 2,
        "fwhm": 3,
        "sky_adu": 4,
        "star_adu": 5,
    }

    def __init__(self, file_path: str) -> None:
        self.__path = Path(file_path)
        self.__is_read = False
        self.__data = None

    def _read(self):
        with self.__path.open() as fd:
            lines = [line.strip() for line in fd.readlines()]
            lines = lines[self.header_rows :]  # Skip headers
            lines = filter(line_str_contains_numbers_and_non_alphabets, lines)
            # Create a 2d list
            lines = [line.split() for line in lines]
            # Convert to 2d numpy array
            self.__data = np.array(lines, dtype="float")
        self.__is_read = True

    def _get_col_value(self, star_no: int, col: str) -> float:
        if not self.__is_read:
            self._read()
        column_number = self.column_numbers[col]
        # Return None if star_no is out of bound
        if star_no < 1 or star_no > len(self.__data):
            return None
        # Note that -1 is necessary bc of 0 indexing
        return self.__data[star_no - 1][column_number]

    def path(self):
        """
        Return the file path
        """
        return self.__path

    def get_x_position_column(self) -> npt.NDArray:
        """
        Returns an ordered array of stars x positions.
        The first row of the array is the x position of star 1, 200th row for star 200,
        and the like
        """
        return self.data()[:, self.column_numbers["x"]]

    def get_y_position_column(self) -> npt.NDArray:
        """
        Returns an ordered array of stars y positions.
        The first row of the array is the y position of star 1, 200th row for star 200,
        and the like
        """
        return self.data()[:, self.column_numbers["y"]]

    def get_star_xy(self, star_no: int) -> Tuple[float]:
        """
        Returns the tuple of `star_no`'s x and y coordinate
        """
        return self._get_col_value(star_no, "x"), self._get_col_value(star_no, "y")

    def get_star_adu(self, star_no: int) -> float:
        """
        Returns the adu for a given star
        """
        return self._get_col_value(star_no, "star_adu")

    def get_star_sky_adu(self, star_no: int) -> float:
        """
        Returns the sky adu for a given star
        """
        return self._get_col_value(star_no, "sky_adu")

    def data(self) -> npt.NDArray:
        """
        Returns the stars data in the reference file as 2 dimensional numpy array
        """
        if not self.__is_read:
            self._read()
        return self.__data

    def __len__(self):
        """
        Returns the number of stars data present in the reference file
        """
        return len(self.data())

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"Reference log file: {self.__path}"

    def create(self, output_path: Path):
        lines = [
            "\n",
            f"Copied from {self.__path}\n",
            "Removed duplicates\n",
            f"Produced on {datetime.now()}\n",
            "Empty lines kept to match format\n" "\n\n\n",
        ]
        with output_path.open("w") as fd:
            fd.writelines(lines)
            header = "%s\t\t%5s\t\t%9s\t\t%4s\t\t%5s\t\t%8s" % (
                "X",
                "Y",
                "Sigma",
                "FWHM",
                "Sky ADU",
                "Star ADU",
            )
            data_fmt = "%4.2f\t\t%4.2f\t\t%2.2f\t\t%2.2f\t\t%4.1f\t\t%8.1f"
            np.savetxt(fd, self.data(), fmt=data_fmt, header=header, comments="")



from typing import Iterable

import numpy as np
import numpy.typing as npt

from m23.trans import createFitFileWithSameHeader


def image_combination(
    images_data: Iterable[npt.NDArray], file_name, fit_file_name_to_copy_header_from
) -> npt.NDArray:
    """
    Combines the `images_data` that's provided and saves the combination in fit file
    that's given by `file_name` copying header information of the fit file from
    `fit_file_name_to_copy_header_from`
    """
    images_data = np.array(images_data)
    combinedImageData = np.sum(images_data, axis=0)

    createFitFileWithSameHeader(
        combinedImageData.astype("int"), file_name, fit_file_name_to_copy_header_from
    )
    return combinedImageData



import logging
from typing import Iterable

import numpy as np

from m23.file.log_file_combined_file import LogFileCombinedFile


def get_star_to_ignore_bit_vector(  # noqa
    log_file_combined_file: LogFileCombinedFile, radius: int
) -> Iterable[int]:
    """
    Looks at the log_file_combined file and returns a bit vector representing
    whether that star should be ignored when calculating the norm factor for the
    night. We create this so that we can ignore stars at the edges which had a
    bright line at the edge after alignment step in old camera images. Note that
    in the bit vector, 0 means that star is to be avoided, 1 means the star is
    to be included.
    """
    y_coordinates = log_file_combined_file.get_x_position_column()
    x_coordinates = log_file_combined_file.get_y_position_column()

    night_date = log_file_combined_file.night_date()
    logger = logging.getLogger("LOGGER_" + str(night_date))

    # We now alter the x and the y values of the stars that don't have
    # any ADU value because we don't want to consider them as the stars in
    # the corners. Note that we reset this value after calculating stars to ignore
    stars_with_bogus_set = {}

    for index in range(len(log_file_combined_file.data())):
        star_no = index + 1
        star_data = log_file_combined_file.get_star_data(star_no)
        if not (
            star_data.sky_adu > 0 and all([adu > 0 for adu in star_data.radii_adu.values()])
        ):  # Note that some may be nan values
            # Set a bogus value on the star's x and y coordinate
            # so that it won't affect corner star calculation
            bogus = 512
            stars_with_bogus_set[star_no] = [x_coordinates[index], y_coordinates[index]]
            logger.debug(
                f"Intranight Linfit. Setting bogus x, y as {bogus} to star {star_no}. "
                + f"Found star data {star_data}. Logfile {log_file_combined_file}"
            )
            x_coordinates[index] = bogus
            y_coordinates[index] = bogus

    bit_vector = []

    # IDL and Python has axes reversed
    dist_from_top_left = np.sqrt(x_coordinates**2 + y_coordinates**2)
    dist_from_top_right = np.sqrt(x_coordinates**2 + (y_coordinates - 1023) ** 2)
    dist_from_bottom_left = np.sqrt((x_coordinates - 1023) ** 2 + y_coordinates**2)
    dist_from_bottom_right = np.sqrt((x_coordinates - 1023) ** 2 + (y_coordinates - 1023) ** 2)

    # indices of stars with least distances from four corners
    min_top_left = np.argmin(dist_from_top_left)
    min_top_right = np.argmin(dist_from_top_right)
    min_bottom_left = np.argmin(dist_from_bottom_left)
    min_bottom_right = np.argmin(dist_from_bottom_right)

    # star coordinates nearest to the four corners
    top_left_star = [x_coordinates[min_top_left], y_coordinates[min_top_left]]
    top_right_star = [x_coordinates[min_top_right], y_coordinates[min_top_right]]
    bottom_left_star = [x_coordinates[min_bottom_left], y_coordinates[min_bottom_left]]
    bottom_right_star = [
        x_coordinates[min_bottom_right],
        y_coordinates[min_bottom_right],
    ]

    # Fit linear lines to the four stars in the four corners, making a quadrilateral
    left_line = np.polyfit(
        [top_left_star[1], bottom_left_star[1]],
        [top_left_star[0], bottom_left_star[0]],
        1,
    )
    right_line = np.polyfit(
        [top_right_star[1], bottom_right_star[1]],
        [top_right_star[0], bottom_right_star[0]],
        1,
    )
    top_line = np.polyfit(
        [top_left_star[1], top_right_star[1]], [top_left_star[0], top_right_star[0]], 1
    )
    bottom_line = np.polyfit(
        [bottom_left_star[1], bottom_right_star[1]],
        [bottom_left_star[0], bottom_right_star[0]],
        1,
    )

    left_line_a, left_line_b = left_line

    def left_line_get_x(y):
        return (y - left_line_b) / left_line_a

    right_line_a, right_line_b = right_line

    def right_line_get_x(y):
        return (y - right_line_b) / right_line_a

    top_line_a, top_line_b = top_line

    def top_line_get_y(x):
        return top_line_a * x + top_line_b

    bottom_line_a, bottom_line_b = bottom_line

    def bottom_line_get_y(x):
        return bottom_line_a * x + bottom_line_b

    def should_include(point):
        y, x = point
        is_between_left_and_right_lines = (
            left_line_get_x(y) + 12 < x and right_line_get_x(y) - 12 > x
        )
        is_between_top_and_bottom = top_line_get_y(x) + 12 < y and bottom_line_get_y(x) - 12 > y
        return is_between_left_and_right_lines and is_between_top_and_bottom

    # We crop in 12 pixels from those four lines, and exclude stars that are
    # outside of this region
    x_and_y_coordinates = np.stack((x_coordinates, y_coordinates), axis=1)
    for star_position in x_and_y_coordinates:
        if should_include(star_position):
            bit_vector.append(1)
        else:
            bit_vector.append(0)

    # Cleanup
    # Reverse the x, y co-ordinates of the stars with where we set bogus co-ordinates
    for star, actual_coordinates in stars_with_bogus_set.items():
        actual_x, actual_y = actual_coordinates
        x_coordinates[star - 1] = actual_x
        y_coordinates[star - 1] = actual_y

    return bit_vector


def is_point_to_left_of_line(a, b, point):
    def eqn_y(x):
        return a * x + b

    def eqn_x(y):
        return (y - b) / a

    x, y = point
    x_prime = eqn_x(y)
    return x_prime > x



import math
from functools import cache
from typing import Dict, Iterable, Tuple

import numpy as np
import numpy.typing as npt

from m23.constants import SKY_BG_BOX_REGION_SIZE
from m23.file.aligned_combined_file import AlignedCombinedFile
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.file.reference_log_file import ReferenceLogFile
from m23.matrix import blockRegions
from m23.utils import half_round_up_to_int


def sky_bg_average_for_all_regions(image_data, region_size):
    """
    Returns a dictionary of background average for all `region_size` sized
    square boxes in `image_data`
    """
    rows, cols = image_data.shape
    no_of_blocks_across_rows = rows // region_size
    no_of_blocks_across_cols = cols // region_size

    # Block in third row first column can be accessed by [2, 0]
    blocks = blockRegions(image_data, (region_size, region_size)).reshape(
        no_of_blocks_across_rows, no_of_blocks_across_cols, region_size, region_size
    )

    # This is a dictionary of background data in all regions The key to this
    # dictionary is the block region number represented as a tuple. For example
    # (1, 2) means second row, third column
    bg_data = {}

    for i in range(no_of_blocks_across_rows):
        for j in range(no_of_blocks_across_cols):
            region = blocks[i][j]
            # Throw out the background of zeroes, since they might be at the edge
            sorted_data = np.sort(region, axis=None)
            non_zero_indices = np.nonzero(sorted_data)
            # Ignore the zeros
            sorted_data = sorted_data[non_zero_indices]

            centered_array = sorted_data[
                int(0.45 * len(sorted_data)) : int(0.55 * len(sorted_data)) + 1
            ]

            # If the data just contains zeros than this will mean that after the
            # zero removal step, our array will be empty and the mean of empty
            # array is nan, we would rather want to write it as 0
            if len(centered_array) == 0:
                bg_data[(i, j)] = 0
            else:
                bg_data[(i, j)] = np.mean(centered_array)

    return bg_data


def extract_stars(
    image_data: npt.NDArray,
    reference_log_file: ReferenceLogFile,
    radii_of_extraction,
    log_file_combined_file: LogFileCombinedFile,
    aligned_combined_file: AlignedCombinedFile,
    date_time_to_use: str = "",
):
    # We save the sky background in each `region_sized`(d)
    # square box of the image in the following
    sky_backgrounds = sky_bg_average_for_all_regions(image_data, SKY_BG_BOX_REGION_SIZE)

    stars_centers_in_new_image = newStarCenters(image_data, reference_log_file)

    star_fluxes = {
        radius: flux_log_for_radius(
            radius, stars_centers_in_new_image, image_data, sky_backgrounds, reference_log_file
        )
        for radius in radii_of_extraction
    }
    no_of_stars = len(star_fluxes[radii_of_extraction[0]])

    log_file_combined_data: LogFileCombinedFile.LogFileCombinedDataType = {}

    for star_no in range(1, no_of_stars + 1):
        weighted_x = stars_centers_in_new_image[star_no - 1][0]
        weighted_y = stars_centers_in_new_image[star_no - 1][1]

        #
        bg_adu_per_pixel = star_fluxes[radii_of_extraction[0]][star_no - 1][1]

        star_FWHM = fwhm(image_data, weighted_x, weighted_y, bg_adu_per_pixel)
        log_file_combined_data[star_no] = LogFileCombinedFile.StarLogfileCombinedData(
            x=weighted_y,  # IDL and Python have Axes reversed
            y=weighted_x,  # Note the axes are reversed by convention
            xFWHM=star_FWHM[1],  # Again, note the axes are reversed by IDL convention
            yFWHM=star_FWHM[0],
            avgFWHM=star_FWHM[2],
            # Note that star_fluxes[radius] is a list of 3 tuples
            # where the elements of the tuple are (total star flux, background
            # flux, subtracted star flux)
            # Also note that we only write sky ADU for one of the radius of extraction
            # This is the usually just the first radius of extraction
            sky_adu=bg_adu_per_pixel,  # Sky ADU from first of extraction
            radii_adu=(
                {radius: star_fluxes[radius][star_no - 1][2] for radius in radii_of_extraction}
            ),
        )
    log_file_combined_file.create_file(
        log_file_combined_data, aligned_combined_file, date_time_to_use
    )


def newStarCenters(imageData, reference_log_file: ReferenceLogFile):
    no_of_stars = len(reference_log_file)

    def centerFinder(star_no):
        x, y = reference_log_file.get_star_xy(star_no)

        colWghtSum = 0
        rowWghtSum = 0
        WghtSum = 0
        for col in range(-5, 6):
            for row in range(-5, 6):
                if math.ceil(math.sqrt((col**2) + (row**2))) <= 5:
                    WghtSum += imageData[half_round_up_to_int(y) + row][
                        half_round_up_to_int(x) + col
                    ]
                    colWghtSum += imageData[half_round_up_to_int(y) + row][
                        half_round_up_to_int(x) + col
                    ] * (x + col)
                    rowWghtSum += imageData[half_round_up_to_int(y) + row][
                        half_round_up_to_int(x) + col
                    ] * (y + row)

        if WghtSum > 0:
            xWght = colWghtSum / WghtSum
            yWght = rowWghtSum / WghtSum
        else:
            xWght = x
            yWght = y

        return yWght, xWght

    return [centerFinder(star_no) for star_no in range(1, no_of_stars + 1)]


def flux_log_for_radius(
    radius: int, stars_center_in_new_image, image_data, sky_backgrounds, ref: ReferenceLogFile
):
    """
    We need to optimize this code to work more efficiently with the caller
    function i.e extract_stars
    """
    regionSize = 64
    pixelsPerStar = np.count_nonzero(circleMatrix(radius))

    def fluxSumForStar(position, radius, star_no) -> Tuple[int]:
        """
        This function returns the flux of of a star at specified `position`
        using `radius` as radius of extraction. Note that this returns a
        three-tuple where the first, second, and third element correspond to
        total star flux, background flux and star flux after background
        subtraction respectively
        """
        # Since we always want a star to fall to the same sky background box, we
        # calculate which star box it falls into by taking the position from the
        # reference file not its weighted X, weighted Y that might change.

        # A better method would be to take the average of surrounding boxes
        # instead because if the sky background noise gets improved because of
        # equipment change, we'll see a step in new data because of how
        # difference of much background we were subtracting previously vs now.

        # Note that the reason y, x are reversed is because our x, y in image
        # data matrix and in the reference file are flipped.
        y_exact, x_exact = ref.get_star_xy(star_no)

        # IDL like round
        x, y = position
        x, y = half_round_up_to_int(x), half_round_up_to_int(y)

        starBox = image_data[x - radius : x + radius + 1, y - radius : y + radius + 1]
        starBox = np.multiply(starBox, circleMatrix(radius))

        backgroundAverageInStarRegion = calculate_star_sky_adu(
            ref.get_star_xy(star_no), sky_backgrounds, box_width=regionSize
        )

        subtractedStarFlux = np.sum(starBox) - backgroundAverageInStarRegion * pixelsPerStar

        # Convert to zero, in case there's any nan.
        # This ensures that two log files correspond to same star number as they are
        # or after reading with something like getLinesWithNumbersFromFile
        # This step makes our normalization code faster than the reslife code written in IDL!
        return (
            np.nan_to_num(np.sum(starBox)),
            np.nan_to_num(backgroundAverageInStarRegion),
            np.nan_to_num(subtractedStarFlux),
        )

    stars_fluxes = [
        fluxSumForStar(position, radius, index + 1)
        for index, position in enumerate(stars_center_in_new_image)
    ]
    return stars_fluxes


@cache
def circleMatrix(radius):
    lengthOfSquare = radius * 2 + 1
    myMatrix = np.zeros(lengthOfSquare * lengthOfSquare).reshape(lengthOfSquare, lengthOfSquare)
    for row in range(-radius, radius + 1):
        for col in range(-radius, radius + 1):
            if math.ceil(math.sqrt((row) ** 2 + (col) ** 2)) <= radius:
                myMatrix[row + radius][col + radius] = 1
    return myMatrix


def fwhm(data, xweight, yweight, aduPerPixel):
    col_sum = 0
    row_sum = 0
    weighted_col_sum = 0
    weighted_row_sum = 0
    for axis in range(-5, 6):
        col_sum += data[half_round_up_to_int(xweight) + axis, half_round_up_to_int(yweight)]
        row_sum += data[half_round_up_to_int(xweight), half_round_up_to_int(yweight) + axis]
        weighted_col_sum += (
            data[half_round_up_to_int(xweight) + axis, half_round_up_to_int(yweight)] - aduPerPixel
        ) * ((half_round_up_to_int(xweight) + axis) - xweight) ** 2
        weighted_row_sum += (
            data[half_round_up_to_int(xweight), half_round_up_to_int(yweight) + axis] - aduPerPixel
        ) * ((half_round_up_to_int(yweight) + axis) - yweight) ** 2
    col_sum = col_sum - (aduPerPixel * 11)
    row_sum = row_sum - (aduPerPixel * 11)

    if weighted_col_sum < 0 or col_sum <= 1:
        xFWHM = 0
    else:
        xFWHM = 2.355 * np.sqrt(weighted_col_sum / (col_sum - 1))
    if weighted_row_sum < 0 or row_sum <= 1:
        yFWHM = 0
    else:
        yFWHM = 2.355 * np.sqrt(weighted_row_sum / (row_sum - 1))
    average_FWHM = np.mean([xFWHM, yFWHM])
    return xFWHM, yFWHM, average_FWHM


def get_star_background_boxes(
    position_in_ref: Tuple[float, float], box_width: float
) -> Iterable[Tuple[int, int]]:
    """
    Returns the sky background boxes that influence the star's brightness.
    @param position_in_ref:
        tuple of x, y position in the reference file. Important to provide position
        in reference and not weighted as these boxes have to be the same for all nights
    @param: width of the sky background box
    returns list of box positions
    """
    threshold = 20
    x, y = position_in_ref

    # Use surrounding boxes for sky background calculation
    b1 = x - threshold, y
    b2 = x + threshold, y
    b3 = x, y + threshold
    b4 = x, y - threshold

    # In order to use just the box that the star falls under for
    # sky background, uncomment the followings:
    # b1 = x, y
    # b2 = x, y
    # b3 = x, y
    # b4 = x, y

    boxes = []
    for box in [b1, b2, b3, b4]:
        bx, by = box
        if bx < 0:
            bx = 0
        if bx >= 1024:
            bx = 1023
        if by < 0:
            by = 0
        if by >= 1024:
            by = 1023
        # Cols are assumed as X, rows Y by IDL convention
        row = by // box_width
        col = bx // box_width
        boxes.append((row, col))
    return boxes


def calculate_star_sky_adu(
    star_position_in_ref: Tuple[float, float],
    sky_backgrounds: Dict[Tuple[int, int], float],
    box_width: int,
) -> float:
    """
    @param star_position_in_ref: x, y position of the star in ref
    @sky_background: Average sky background per pixel in all different sky
    background boxes

    Return the average star background per pixel for the star
    """
    boxes = get_star_background_boxes(star_position_in_ref, box_width)
    sky_bg_in_boxes = list(map(sky_backgrounds.get, boxes))
    return np.mean(sky_bg_in_boxes)



from pathlib import Path

from m23.calibrate.master_calibrate import makeMasterDark, makeMasterFlat
from m23.constants import INPUT_CALIBRATION_FOLDER_NAME
from m23.file.masterflat_file import MasterflatFile
from m23.matrix import crop
from m23.processor.generate_masterflat_config_loader import (
    MasterflatGeneratorConfig,
    validate_generate_masterflat_config_file,
)
from m23.utils import (
    fit_data_from_fit_images,
    get_darks,
    get_date_from_input_night_folder_name,
    get_flats,
    sorted_by_number,
)


def generate_masterflat_auxiliary(config: MasterflatGeneratorConfig) -> None:
    """
    Generates masterflat based on the configuration provided
    this function assumes that the configuration provided is valid as it
    should be only calledfrom generate_master_flat that checks for the validity
    of the configuration file before calling this function.
    """
    rows, cols = config["image"]["rows"], config["image"]["columns"]
    crop_region = config["image"].get("crop_region", [])
    NIGHT_INPUT_CALIBRATION_FOLDER = config["input"] / INPUT_CALIBRATION_FOLDER_NAME
    # Note the order is important when generating masterflat
    flats = fit_data_from_fit_images(sorted_by_number(get_flats(NIGHT_INPUT_CALIBRATION_FOLDER)))
    darks = fit_data_from_fit_images(get_darks(NIGHT_INPUT_CALIBRATION_FOLDER))
    night_date = get_date_from_input_night_folder_name(config["input"])

    # Crop images if crop region is defined
    if len(crop_region) > 0:
        flats = [crop(matrix, rows, cols) for matrix in flats]

    # We have to first create master dark before creating masterflat
    # as masteflat requires masterdark. Note that we're passing saveAs
    # as None because we don't want to save the masterdark created in this
    # process
    masterDarkData = makeMasterDark(
        listOfDarkData=darks,
    )

    # Make master flat
    filename = MasterflatFile.generate_file_name(night_date)
    save_file_path = config["output"] / filename

    makeMasterFlat(
        saveAs=save_file_path,
        masterDarkData=masterDarkData,
        headerToCopyFromName=next(
            get_flats(NIGHT_INPUT_CALIBRATION_FOLDER)
        ).absolute(),  # Gets absolute path of first flat file,
        listOfFlatData=flats,
    )


def generate_masterflat(file_path: str):
    """
    Starts generating masterflat based on the configuration specified in the
    file given by `file_path`
    """
    validate_generate_masterflat_config_file(
        Path(file_path), on_success=generate_masterflat_auxiliary
    )



import csv
import random

from m23.file.color_normalized_file import ColorNormalizedFile
from m23.processor.nights_csv_config_loader import (
    NightsCSVConfig,
    validate_nights_csv_config_file,
)


def create_nights_csv_auxiliary(config: NightsCSVConfig):
    """
    Creates and saves a csv of star fluxes for night specified in the
    contents of the `file_path`.
    """
    radius = config["radius"]
    fluxes_file_name = f"fluxes_{radius}px.txt"
    output_folder = config["output"]
    fluxes_file = output_folder / fluxes_file_name
    # We ensure that the filename doesn't already exist so that we don't override
    # existing file
    while fluxes_file.exists():
        fluxes_file = output_folder / f"flux_{radius}px_{random.randrange(1, 100)}.txt"
    color_normalized_files = [
        ColorNormalizedFile(file) for file in config["color_normalized_files"]
    ]
    color_normalized_files = sorted(color_normalized_files, key=lambda x: x.night_date())

    all_color_files_data = [
        color_normalized_files[i].data() for i in range(len(color_normalized_files))
    ]
    night_dates = [
        str(color_normalized_files[i].night_date()) for i in range(len(color_normalized_files))
    ]

    with open(fluxes_file, "w", newline="") as file:
        writer = csv.writer(file, delimiter=",")
        writer.writerow(["Star#"] + night_dates)
        for star_no in range(len(all_color_files_data[0])):  # Writes the data of 2510 stars
            star_data = [
                all_color_files_data[i][star_no + 1].normalized_median_flux
                for i in range(len(color_normalized_files))
            ]
            writer.writerow([str(star_no + 1)] + star_data)
        if (
            len(all_color_files_data[0]) == 2508
        ):  # If no data for stars 2509 and 2510, then write them as empty
            writer.writerow(["2509"] + ["0" for i in range(len(color_normalized_files))])
            writer.writerow(["2510"] + ["0" for i in range(len(color_normalized_files))])
    file.close()

    # Create a combined sky background file
    sky_bg_file = output_folder / "sky_bg.txt"
    while sky_bg_file.exists():
        sky_bg_file = output_folder / f"sky_bg_{random.randrange(1, 100)}.txt"

    sky_bg_files = config["sky_bg_files"]
    if len(sky_bg_files) == 0:
        raise Exception("No sky background files found.")

    with open(sky_bg_files[0], "r") as fd:
        sky_bg_header = fd.readline()
    sky_bg_header_cols = len(sky_bg_header.split())

    lines_to_write = [sky_bg_header]
    for f in sky_bg_files:
        with open(f, "r") as fd:
            lines = fd.readlines()[1:]
            if length := len(lines):
                line_to_write = lines[length // 2]
                if len(line_to_write.split()) != sky_bg_header_cols:
                    raise Exception(
                        f"file {f} and {sky_bg_files[0]} have different number of columns"
                    )
                lines_to_write.extend(lines[1:])

    with open(sky_bg_file, "w") as fd:
        fd.writelines(lines_to_write)


def create_nights_csv(file_path: str):
    """
    Creates and saves a csv of star fluxes for night specified in the
    contents of the `file_path`. This function calls
    `validate_nights_csv_config_file` to do most of its job
    """
    validate_nights_csv_config_file(file_path, create_nights_csv_auxiliary)



import os
import re
from datetime import date, datetime
from decimal import ROUND_HALF_UP, Decimal
from pathlib import Path, PosixPath
from typing import Iterable, List, Union

import numpy as np
from astropy.io.fits import getdata as getfitsdata
from numpy.typing import DTypeLike

from m23.constants import (
    INPUT_NIGHT_FOLDER_NAME_DATE_FORMAT,
    OUTPUT_NIGHT_FOLDER_NAME_DATE_FORMAT,
)
from m23.file.raw_image_file import RawImageFile

# local imports
from .rename import rename


def get_image_number_in_log_file_combined_file(file: Path) -> int:
    """
    Returns the image number log file combined file, or raises error if image number not found
    Examples:
        In the filename, 07-07-18_m23_7.0-112.txt, the image number is 112
    """
    results = re.findall(r"^.*-(\d+)\.txt", file.name)
    if len(results) == 0:
        raise ValueError(f"{file.name} is not in something-xxx.txt format")
    else:
        return int(results[0])


def get_image_number_in_fit_file(file: Path) -> int:
    """
    Returns the image number of the fit file, or raises error if image number not found
    Examples:
        In the filename, m23_7.0-010.fit, image number is 10
        More generally, In something-xxx.fit, integer representing xxx defines the image number
    """
    results = re.findall(r"^.*-(\d+)\.fit", file.name)
    if len(results) == 0:
        raise ValueError(f"{file.name} is not in something-xxx.fit format")
    else:
        return int(results[0])


def get_flats(folder: Path) -> Iterable[PosixPath]:
    """
    Return a list of flat files in `folder` provided
    """
    return folder.glob("*flat*.fit")


def get_darks(folder: Path) -> Iterable[PosixPath]:
    """
    Return a list of dark files in `folder` provided
    """
    return folder.glob("*dark*.fit")


def get_all_fit_files(folder: Path) -> Iterable[PosixPath]:
    """
    Return a list of all fit files in `folder` provided
    """
    return folder.glob("*.fit")


def get_raw_images(folder: Path) -> Iterable[RawImageFile]:
    """
    Return a list `RawImageFile` files in `folder` provided sorted asc. by image number
    Note that only filenames matching the naming convention of RawImageFile are returned
    """
    all_files = [RawImageFile(file.absolute()) for file in folder.glob("*.fit")]
    # Filter files whose filename don't match naming convention
    all_files = filter(lambda raw_image_file: raw_image_file.is_valid_file_name(), all_files)
    # Sort files by image number
    return sorted(all_files, key=lambda raw_image_file: raw_image_file.image_number())


def time_taken_to_capture_and_save_a_raw_file(folder_path: Path) -> int:
    """
    Returns the average time taken to capture the raw image.  Note that this
    may be different from the `image_duration` which is the time of camera
    exposure. This because it also takes some time to save the fit image.
    This function looks at the datetime of the first and the last raw image in
    `folder_path` and calculates the average time taken for an image.

    Raises
        Exception if no raw image is present in the given folder

    """
    raw_images: Iterable[RawImageFile] = list(get_raw_images(folder_path))
    first_img = raw_images[0]
    last_image = raw_images[-1]
    no_of_images = len(raw_images)
    duration = (last_image.datetime() - first_img.datetime()).seconds
    return duration / no_of_images


def get_radius_folder_name(radius: int) -> str:
    """
    Returns the folder name to use for a given radius pixel of extraction
    """
    radii = {
        1: "One",
        2: "Two",
        3: "Three",
        4: "Four",
        5: "Five",
        6: "Six",
        7: "Seven",
        8: "Eight",
        9: "Nine",
    }
    if result := radii.get(radius):
        return f"{result} Pixel Radius"
    else:
        return f"{radius} Pixel Radius"


def get_date_from_input_night_folder_name(name: str | Path) -> date:
    if issubclass(type(name), Path):
        name = name.name
    return datetime.strptime(name, INPUT_NIGHT_FOLDER_NAME_DATE_FORMAT).date()


def get_output_folder_name_from_night_date(night_date: date) -> str:
    return night_date.strftime(OUTPUT_NIGHT_FOLDER_NAME_DATE_FORMAT)


def fitFilesInFolder(folder, fileType="All"):
    allFiles = os.listdir(folder)
    fileType = fileType.lower()

    allFiles = list(filter(lambda x: x.endswith(".fit"), allFiles))
    if fileType == "all":
        return allFiles
    else:
        return list(filter(lambda x: x.__contains__(fileType), allFiles))


def fitDataFromFitImages(images):
    return [getfitsdata(item) for item in images]


def fit_data_from_fit_images(images: Iterable[str | Path]) -> List[DTypeLike]:
    return [getfitsdata(item) for item in images]


def get_log_file_name(night_date: date):
    return f"Night-{night_date}-Processing-log.txt"


def half_round_up_to_int(num: float):
    # Python and IDL round up half numbers differently
    # In python round(1.5) is 2 while round(2.5) is 2
    # while in IDL all half numbers are rounded up
    # This function mimics IDL behaviour
    return int(Decimal(num).to_integral_value(rounding=ROUND_HALF_UP))


def customMedian(arr, *args, **kwargs):
    """
    Median similar to the default version of IDL Median
    https://github.com/LutherAstrophysics/python-helpers/issues/8
    """
    arr = np.array(arr)
    if len(arr) % 2 == 0:
        newArray = np.append(arr, [np.multiply(np.ones(arr[0].shape), np.max(arr))], axis=0)
        return np.median(newArray, *args, **kwargs)
    else:
        return np.median(arr, *args, **kwargs)


def sorted_by_number(lst: Iterable[Union[str, Path]]) -> Iterable[Union[str, Path]]:
    """
    Returns sorted list of `lst` where sorting is done in ascending order by the
    the first number present in the string. If no string is present, falls back
    to alphabetic sorting
    """
    # Note that this step is important if the `lst` passed is a generator
    # This is because, after we've read it once, it would be empty
    lst = list(lst)
    if not all([isinstance(x, str) or isinstance(x, Path) for x in lst]):
        raise ValueError("items of list must be either str or Path instance")
    # Get filename if path objects are given
    lst_path_normalized = [x if isinstance(x, str) else x.name for x in lst]
    # Sort by alphabet (secondary)
    alphabet_sorted = sorted(enumerate(lst_path_normalized), key=lambda x: x[1])
    # Sort by first number present in the
    MATCHER = re.compile(r"\D*(\d*).*")
    foo = zip(*list(sorted(alphabet_sorted, key=lambda x: int(MATCHER.match(x[1])[1] or 0))))
    indices, _ = foo
    return [lst[i] for i in indices]


__all__ = [
    "customMedian",
    "fitFilesInFolder",
    "rename",
    "get_closet_date",
    "raw_data_name_format",
    "sorted_by_number",
]



import logging
from datetime import timedelta
from pathlib import Path

import numpy as np
from logger_tt import setup_logging
from m23.align import image_alignment
from m23.calibrate.calibration import calibrateImages
from m23.constants import (ALIGNED_COMBINED_FOLDER_NAME, ALIGNED_FOLDER_NAME,
                           LOG_FILES_COMBINED_FOLDER_NAME,
                           M23_RAW_IMAGES_FOLDER_NAME,
                           RAW_CALIBRATED_FOLDER_NAME)
from m23.exceptions import CouldNotAlignException
from m23.extract import extract_stars
from m23.file.aligned_combined_file import AlignedCombinedFile
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.file.raw_image_file import RawImageFile
from m23.file.reference_log_file import ReferenceLogFile
from m23.matrix import crop
from m23.matrix.fill import fillMatrix
from m23.processor.config_loader import Config, ConfigInputNight
from m23.utils import time_taken_to_capture_and_save_a_raw_file

setup_logging(use_multiprocessing=True)

def align_combined_extract(
    config: Config,
    night: ConfigInputNight,
    output: Path,
    night_date,
    nth_combined_image,
    raw_images,
    master_dark_data,
    master_flat_data,
    alignment_stats_file,
    image_duration,
    log_files_to_normalize_queue,
):
    logger = logging.getLogger("LOGGER_" + str(night_date))

    # Define relevant input folders for the night being processed
    NIGHT_INPUT_FOLDER: Path = night["path"]
    NIGHT_INPUT_IMAGES_FOLDER = NIGHT_INPUT_FOLDER / M23_RAW_IMAGES_FOLDER_NAME

    # Define and create relevant output folders for the night being processed
    JUST_ALIGNED_NOT_COMBINED_OUTPUT_FOLDER = output / ALIGNED_FOLDER_NAME
    ALIGNED_COMBINED_OUTPUT_FOLDER = output / ALIGNED_COMBINED_FOLDER_NAME
    LOG_FILES_COMBINED_OUTPUT_FOLDER = output / LOG_FILES_COMBINED_FOLDER_NAME
    RAW_CALIBRATED_OUTPUT_FOLDER = output / RAW_CALIBRATED_FOLDER_NAME

    ref_image_path = config["reference"]["image"]
    ref_file_path = config["reference"]["file"]
    reference_log_file = ReferenceLogFile(ref_file_path)
    rows, cols = config["image"]["rows"], config["image"]["columns"]
    no_of_images_to_combine = config["processing"]["no_of_images_to_combine"]

    crop_region = config["image"]["crop_region"]
    save_aligned_images = config["output"]["save_aligned"]
    save_calibrated_images = config["output"]["save_calibrated"]
    radii_of_extraction = config["processing"]["radii_of_extraction"]

    from_index = nth_combined_image * no_of_images_to_combine
    # Note the to_index is exclusive
    to_index = (nth_combined_image + 1) * no_of_images_to_combine

    # NOTE
    # It's very easy to get confused between no_of_combined_images
    # and the no_of_images_to_combine. THe later is the number of raw images
    # that are combined together to form on aligned combined image

    images_data = [raw_image_file.data() for raw_image_file in raw_images[from_index:to_index]]
    # Ensure that image dimensions are as specified by rows and cols
    # If there's extra noise cols or rows, we crop them
    images_data = [crop(matrix, rows, cols) for matrix in images_data]

    # Calibrate images
    images_data = calibrateImages(
        masterDarkData=master_dark_data,
        masterFlatData=master_flat_data,
        listOfImagesData=images_data,
    )

    if save_calibrated_images:
        for index, raw_image_index in enumerate(range(from_index, to_index)):
            raw_img = raw_images[raw_image_index]
            calibrated_image = RawImageFile(RAW_CALIBRATED_OUTPUT_FOLDER / raw_img.path().name)
            calibrated_image.create_file(images_data[index], raw_img)
            logger.info(f"Saving calibrated image. {raw_image_index}")

    # Fill out the cropped regions with value of 1
    # Note, it's important to fill after the calibration step
    if len(crop_region) > 0:
        images_data = [fillMatrix(matrix, crop_region, 1) for matrix in images_data]

    # Alignment
    # We want to discard this set of images if any one image in this set cannot be aligned
    aligned_images_data = []
    for index, image_data in enumerate(images_data):
        raw_image_to_align = raw_images[from_index + index]
        raw_image_to_align_name = raw_image_to_align.path().name
        try:
            aligned_data, statistics = image_alignment(image_data, ref_image_path)
            aligned_images_data.append(aligned_data)
            # We add the transformation statistics to the alignment stats
            # file Information of the file that can't be aligned isn't
            # written only in the logfile. This is intended so that we can
            # easily process the alignment stats file if we keep it in a TSV
            # like format

            # Note that we're down-scaling the matrix dtype from float to int32 for
            # support in the image viewing softwares. For the combination step though
            # we are using the more precise float data. This means that if you read
            # the data of the aligned images from the fit file and combined them yourself
            # that is going to be off by a small amount that the data in the aligned
            # combined image.
            aligned_image = RawImageFile(
                JUST_ALIGNED_NOT_COMBINED_OUTPUT_FOLDER / raw_image_to_align_name
            )

            if save_aligned_images:
                aligned_image.create_file(aligned_data.astype("int32"), raw_image_to_align)

            alignment_stats_file.add_record(raw_image_to_align_name, statistics)
            logger.info(f"Aligned {raw_image_to_align_name}")
        except CouldNotAlignException:
            logger.error(f"Could not align image {raw_image_to_align}")
            logger.error(f"Skipping combination {from_index}-{to_index}")
            break

    # We proceed to next set of images if the alignment wasn't successful for any one
    # image in the combination set. We now this by checking no of aligned images.
    if len(aligned_images_data) < no_of_images_to_combine:
        return

    # Combination
    combined_images_data = np.sum(aligned_images_data, axis=0)

    # We take the middle image from the combination as the sample This is
    # the image whose header will be copied to the combined image fit file
    midpoint_index = from_index + no_of_images_to_combine // 2
    sample_raw_image_file = raw_images[midpoint_index]

    aligned_combined_image_number = to_index // no_of_images_to_combine
    aligned_combined_file_name = AlignedCombinedFile.generate_file_name(
        image_duration, aligned_combined_image_number
    )
    aligned_combined_file = AlignedCombinedFile(
        ALIGNED_COMBINED_OUTPUT_FOLDER / aligned_combined_file_name
    )
    # Image viewing softwares like Astromagic and Fits Liberator don't work
    # if the image data type is float, for some reason that we don't know.
    # So we're setting the datatype to int32 which has enough precision for
    # us.
    aligned_combined_file.create_file(combined_images_data.astype("int32"), sample_raw_image_file)
    logger.info(f"Combined images {from_index}-{to_index}")

    # Extraction
    log_file_combined_file_name = LogFileCombinedFile.generate_file_name(
        night_date, aligned_combined_image_number, image_duration
    )
    log_file_combined_file = LogFileCombinedFile(
        LOG_FILES_COMBINED_OUTPUT_FOLDER / log_file_combined_file_name
    )

    date_time_to_use = get_datetime_to_use(
        aligned_combined_file, night, no_of_images_to_combine, NIGHT_INPUT_IMAGES_FOLDER
    )

    extract_stars(
        combined_images_data,
        reference_log_file,
        radii_of_extraction,
        log_file_combined_file,
        aligned_combined_file,
        date_time_to_use,
    )

    log_files_to_normalize_queue.put(log_file_combined_file)
    logger.info(f"Extraction from combination {from_index}-{to_index} completed")


def get_datetime_to_use(
    aligned_combined: AlignedCombinedFile,
    night_config: ConfigInputNight,
    no_of_raw_images_in_one_combination: int,
    raw_images_folder: Path,
) -> str:
    """
    Returns the datetime to use in the logfile combined file,
    based on the a given `config` and `aligned_combine` file

    Returns an empty string if no datetime is available to use
    """

    # We use the same format of the datetime string as is in the
    # header of our fit files
    datetime_format = aligned_combined.date_observed_datetime_format

    # If the datetime option was passed in the header we use that one
    # Otherwise we use the datetime in the header, if that's present
    if start := night_config.get("starttime"):
        duration_of_raw_img = time_taken_to_capture_and_save_a_raw_file(raw_images_folder)
        img_no = aligned_combined.image_number()
        time_taken_to_capture_one_combined_image = (
            duration_of_raw_img * no_of_raw_images_in_one_combination
        )
        seconds_elapsed_from_beginning_of_night = (
            time_taken_to_capture_one_combined_image * (img_no - 1)
            + time_taken_to_capture_one_combined_image * 0.5
        )
        return (start + timedelta(seconds=seconds_elapsed_from_beginning_of_night)).strftime(
            datetime_format
        )
    elif datetime_in_aligned_combined := aligned_combined.datetime():
        return datetime_in_aligned_combined.strftime(datetime_format)
    else:
        return ""



import datetime
import re
from pathlib import Path

import numpy.typing as npt
from astropy.io import fits
from astropy.io.fits.header import Header

from m23.constants import OBSERVATION_DATETIME_FORMAT
from m23.file.raw_image_file import RawImageFile


class AlignedCombinedFile:
    # Class attributes
    file_name_re = re.compile(r"m23_(\d+\.?\d*)[-_](\d+).fit")
    date_observed_header_name = "DATE-OBS"
    time_observed_header_name = "TIME-OBS"
    date_observed_datetime_format = OBSERVATION_DATETIME_FORMAT

    @classmethod
    def generate_file_name(cls, img_duration: float, img_number: int) -> str:
        """
        Generates filename based on the given `img_duration` and `img_number`
        """
        return f"m23_{img_duration}-{img_number:04}.fit"

    def __init__(self, file_path) -> None:
        self.__path = Path(file_path)
        self.__is_read = False
        self.__data = None
        self.__header = None

    def _read(self):
        if not self.exists():
            raise FileNotFoundError(f"File not found {self.path()}")
        if not self.path().is_file() and self.__path.suffix != ".fit":
            raise ValueError(f"Invalid fit file {self.path()}")
        with fits.open(self.path()) as fd:
            self.__header = fd[0].header
            self.__data = fd[0].data
        self.__is_read = True

    def exists(self):
        return self.path().exists()

    def datetime(self) -> None | datetime.datetime:
        """
        Returns the datetime object of the time observed. Parses the datetime
        field from the header of the image
        """
        timestr = self.header().get(self.time_observed_header_name)
        datestr = self.header().get(self.date_observed_header_name)
        if timestr:
            # If time header is present, construct the datetime from date and time header
            return datetime.datetime.strptime(
                f"{datestr}T{timestr}", self.date_observed_datetime_format
            )
        else:
            # If time no time header is present, we assume that both date and
            # time are present in the date header in format given by
            # `self.date_observed_datetime_format`
            return datetime.datetime.strptime(f"{datestr}", self.date_observed_datetime_format)

    def is_valid_file_name(self):
        """
        Checks if the file name is valid as per the file naming conventions
        of m23 data processing library.
        """
        return bool(self.file_name_re.match(self.path().name))

    def image_duration(self):
        """
        Returns the image duration if the filename is valid
        """
        if not self.is_valid_file_name():
            raise ValueError(f"{self.path().name} doesn't match naming conventions")
        return float(self.file_name_re.match(self.path().name)[1])

    def image_number(self):
        """
        Returns the image number if the filename is valid
        """
        if not self.is_valid_file_name():
            raise ValueError(f"{self.path().name} doesn't match naming conventions")
        return int(self.file_name_re.match(self.path().name)[2])

    def data(self) -> npt.NDArray:
        if not self.__is_read:
            self._read()
        return self.__data

    def header(self) -> Header:
        if not self.__is_read:
            self._read()
        return self.__header

    def path(self):
        return self.__path

    def create_file(self, data: npt.NDArray, copy_header_from: RawImageFile) -> None:
        """
        Create a fit file based on provided np array `data`.
        It copies the header information from the `RawImageFile`
        """
        fits.writeto(self.path(), data, header=copy_header_from.header())

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"Aligned combined file: {self.path()}"



from datetime import datetime


def get_closet_date(base_date: str, list_of_dates: str, format: str):
    """
    From the list of dates, chooses a date that is closest to
    a particular date.

    param base_date : The base date from which to evaluate distance of other dates
    param list_of_dates : Dates to choose from
    format: The string format in which dates are present
    return : an item from list_of_dates
    """

    # Raise error is the list_of_dates is empty
    if len(list_of_dates) == 0:
        raise ValueError("No dates to choose from")

    def make_datetime(x: str):
        return datetime.strptime(x, format)

    base_datetime = make_datetime(base_date)
    list_of_datetimes = list(map(make_datetime, list_of_dates))

    list_of_datetimes.sort(key=lambda x: abs((base_datetime - x)).days)
    return list_of_datetimes[0].strftime(format)


raw_data_name_format = "%B %d, %Y"



import matplotlib

# Use non interactive backend
matplotlib.use("agg")

import itertools  # noqa
import sys
from pathlib import Path
from typing import Callable, Dict, Iterable

import numpy as np
from matplotlib import pyplot as plt

from m23.constants import CHARTS_FOLDER_NAME, FLUX_LOGS_COMBINED_FOLDER_NAME
from m23.file.log_file_combined_file import LogFileCombinedFile
from m23.file.normfactor_file import NormfactorFile
from m23.utils import get_date_from_input_night_folder_name, get_radius_folder_name


def draw_normfactors_chart(
    log_files_used: Iterable[LogFileCombinedFile],
    night_folder: Path,
    radii_of_extraction: Iterable[int],
) -> None:
    """
    Draws normfactors vs image number chart for for a provided `night_folder`
    Note that you must also provided `log_files_used` because otherwise there is no way to know
    which logfile corresponds to which norm factor

    param: log_files_used: The list or sequence of log files to used when doing
            intra night normalization
    param: night_folder: The night folder that hosts other folders like Flux Logs Combined, etc.
    return: None

    Side effects:
    Creates a charts folder in night_folder and saves the normfactors charts there
    """
    # Sort log files
    log_files_used.sort(key=lambda logfile: logfile.img_number())
    flux_log_combined_folder = night_folder / FLUX_LOGS_COMBINED_FOLDER_NAME
    night_date = get_date_from_input_night_folder_name(night_folder.name)
    chart_folder = night_folder / CHARTS_FOLDER_NAME
    chart_folder.mkdir(parents=True, exist_ok=True)  # Create folder if it doesn't exist
    valid_folder_names = list(map(get_radius_folder_name, radii_of_extraction))

    for radius_folder in flux_log_combined_folder.glob("*Radius*"):
        if radius_folder.name not in valid_folder_names:
            continue
        normfactor_files = list(radius_folder.glob("*normfactor*"))
        if len(normfactor_files) != 1:
            sys.stderr.write(
                f"Expected to find 1 normfactor file, found {len(normfactor_files)} in {radius_folder}\n"  # noqa
            )
            # Skip this radius folder
            continue
        normfactor_file = NormfactorFile(normfactor_files[0].absolute())
        normfactor_data = normfactor_file.data()
        log_file_number_to_normfactor_map = {}
        if len(log_files_used) != len(normfactor_data):
            sys.stderr.write(
                "Make sure you're providing exactly the same number of logfiles as there are normfactor values\n"  # noqa
            )
            raise ValueError(
                f"""Mismatch between number of logfiles and the normfactors.
                Found logfiles: {len(log_files_used)}
                Found normfactors: {len(normfactor_data)}
                """
            )
        for index, log_file in enumerate(log_files_used):
            log_file_number_to_normfactor_map[log_file.img_number()] = normfactor_data[index]

        first_img_number = log_files_used[0].img_number()
        last_img_number = log_files_used[-1].img_number()
        chart_name = f"{night_date} normfactors_chart_{first_img_number}-{last_img_number}_{radius_folder.name}.png"  # noqa
        chart_file_path = chart_folder / chart_name
        x, y = zip(
            *log_file_number_to_normfactor_map.items()
        )  # Unpack a list of pairs into two tuples
        plt.figure(dpi=300)
        plt.plot(x, y, "b+")
        plt.xlabel("Log file number")
        plt.ylabel("Normfactor")
        plt.grid()
        plt.title(f"{night_date}")
        plt.savefig(chart_file_path)
        plt.close()  # Important to close the figure


def draw_internight_color_chart(
    night: Path,
    radius: int,
    section_x_values: Dict[int, Iterable],
    section_y_values: Dict[int, Iterable],
    section_color_fit_fn: Dict[int, Callable[[float], float]],
) -> Dict[int, float]:
    """
    Creates and saves color chart for a given night for a particular radius data
    Returns the median value of each section of the chart
    """
    chart_folder = night / CHARTS_FOLDER_NAME
    night_date = get_date_from_input_night_folder_name(night.name)
    chart_name = f"{night_date} Color Curve Fit Radius {radius}_px.png"
    chart_save_path = chart_folder / chart_name
    sections = sorted(section_x_values.keys())
    all_x_values = list(
        itertools.chain.from_iterable([section_x_values[section] for section in sections])
    )
    all_y_values = list(
        itertools.chain.from_iterable([section_y_values[section] for section in sections])
    )
    plt.figure(dpi=300, figsize=(10, 6))
    plt.rcParams["axes.facecolor"] = "black"
    plt.plot(all_x_values, all_y_values, "wo", ms=0.5)
    # Plot the curves for each of the three sections
    section_line_colors = ["blue", "green", "red"]
    median_of_sections = {}
    for index, section in enumerate(sections):
        x = section_x_values[section]
        x_min = np.min(x)
        x_max = np.max(x)
        x_new = np.linspace(x_min, x_max, 300)
        y_new = [section_color_fit_fn[section](i) for i in x_new]
        # Add the median value for each section to the dictionary
        median_of_sections[index] = np.median(y_new)
        plt.plot(x_new, y_new, color=section_line_colors[index], linewidth=2)
    ax = plt.gca()
    ax.set_xlim([0, np.max(all_x_values) + 3 * np.std(all_x_values)])
    ax.set_ylim([0, np.max(all_y_values) + 3 * np.std(all_y_values)])
    plt.title(f"{night_date}")
    plt.xlabel("Color")
    plt.ylabel("Flux Ratio")
    plt.savefig(chart_save_path)
    plt.close()  # Important to close the figure
    return median_of_sections


def draw_internight_brightness_chart(
    night: Path,
    radius: int,
    section_x_values: Dict[int, Iterable],
    section_y_values: Dict[int, Iterable],
    section_fit_fn: Dict[int, Callable[[float], float]],
) -> Dict[int, float]:
    """
    Creates and saves brightness chart for a given night for a particular radius data
    Returns the median value of each section of the chart
    """
    chart_folder = night / CHARTS_FOLDER_NAME
    night_date = get_date_from_input_night_folder_name(night.name)
    chart_name = f"{night_date} Brightness Curve Fit Radius {radius}_px.png"
    chart_save_path = chart_folder / chart_name
    sections = sorted(section_x_values.keys())
    all_x_values = list(
        itertools.chain.from_iterable([section_x_values[section] for section in sections])
    )
    all_y_values = list(
        itertools.chain.from_iterable([section_y_values[section] for section in sections])
    )
    plt.figure(dpi=300, figsize=(10, 6))
    plt.rcParams["axes.facecolor"] = "black"
    plt.plot(all_x_values, all_y_values, "wo", ms=0.5)
    # Plot the curves for each of the three sections
    section_line_colors = ["blue", "green", "red"]
    median_of_sections = {}
    for index, section in enumerate(sections):
        x = section_x_values[section]
        x_min = np.min(x)
        x_max = np.max(x)
        x_new = np.linspace(x_min, x_max, 300)
        y_new = [section_fit_fn[section](i) for i in x_new]
        median_of_sections[index] = np.median(y_new)
        plt.plot(x_new, y_new, color=section_line_colors[index], linewidth=2)

    plt.title(f"{night_date}")
    plt.xlabel("Magnitudes")
    plt.ylabel("Flux Ratio")
    plt.savefig(chart_save_path)
    plt.close()  # Important to close the figure
    return median_of_sections



from pathlib import Path
from typing import Callable, List, TypedDict

import toml

from m23.constants import COLOR_NORMALIZED_FOLDER_NAME, SKY_BG_FOLDER_NAME
from m23.file.color_normalized_file import ColorNormalizedFile
from m23.file.sky_bg_file import SkyBgFile
from m23.utils import get_date_from_input_night_folder_name, get_radius_folder_name


class NightsCSVConfig(TypedDict):
    color_normalized_files: List[Path]
    sky_bg_files: List[Path]
    output: Path
    radius: int


def handle_night(night, radius, color_normalized_files, sky_bg_files):
    night_path = Path(night)
    if not night_path.exists():
        raise Exception(f"night {night} doesn't exist")
    try:
        night_date = get_date_from_input_night_folder_name(night_path)
    except ValueError:
        raise Exception(f"{night} name doesn't match naming conventions")
    color_normalized_file_name = ColorNormalizedFile.get_file_name(night_date, radius)
    sky_bg_file_name = SkyBgFile.generate_file_name(night_date)
    cn_file_path = (
        night_path
        / COLOR_NORMALIZED_FOLDER_NAME
        / get_radius_folder_name(radius)
        / color_normalized_file_name
    )
    sky_bg_file_path = night_path / SKY_BG_FOLDER_NAME / sky_bg_file_name
    if not sky_bg_file_path.exists():
        raise Exception(
            f"Sky background file for {night_date} doesn't exist with name {sky_bg_file_path.name}"
            f" in {night_path / SKY_BG_FOLDER_NAME}"
        )
    if not cn_file_path.exists():
        raise Exception(
            f"Color normalized file for {night_date} doesn't exist with name {cn_file_path.name}"
        )
    color_normalized_files.append(cn_file_path)
    sky_bg_files.append(sky_bg_file_path)


def validate_nights_csv_config_file(
    file_path: Path, on_success: Callable[[NightsCSVConfig], None]
):
    """
    This method reads configuration file for generating a csv file of stars
    daily flux values for the given nights and calls the unary on_success
    function with the configuration file if the config file is valid
    """
    if not file_path.exists():
        raise FileNotFoundError("Cannot find configuration file")
    # Make sure the output path exists
    config_data = toml.load(file_path)
    input_nights = config_data.get("input")
    radius = config_data.get("radius")

    # Radius is an int
    if type(radius) != int:
        raise Exception("Improper radius value, must be an int")

    # Input key exists
    if not input_nights:
        raise Exception("Can't find 'input' in config file")
    # All input locations are valid
    if type(input_nights) != list:
        raise Exception("'input' must be a list of night paths")

    color_normalized_files = []
    sky_bg_files = []

    for night in input_nights:
        handle_night(night, radius, color_normalized_files, sky_bg_files)

    output_location = config_data.get("output")
    # Output key exists
    if not output_location:
        raise Exception("Can't find 'output' in config file")
    # Output dir exists
    output_path = Path(output_location)
    if not output_path.is_dir() or not output_path.exists():
        raise Exception("Output path must be an existing directory")

    on_success(
        {
            "color_normalized_files": color_normalized_files,
            "sky_bg_files": sky_bg_files,
            "radius": radius,
            "output": output_path,
        }
    )



import numpy as np

from m23.trans import createFitFileWithSameHeader
from m23.utils import customMedian, fitDataFromFitImages, sorted_by_number

# This code is a direct implementation of steps
# mentioned in Handbook of Astronomical Image `Processing by
# Richard Berry and James Burnell version 2.0 section 6.3 Standard Calibration


#  makeMasterBias
#
#  purpose: creates masterBias, saves to fileName + returns masterBiasData
#
def makeMasterBias(saveAs, headerToCopyFromName=None, listOfBiasNames=None, listOfBiasData=None):
    if listOfBiasNames:
        listOfBiasData = fitDataFromFitImages(listOfBiasNames)

    if not listOfBiasNames and not listOfBiasData:
        raise Exception("Neither Bias data nor names were provided")

    if not headerToCopyFromName and listOfBiasNames:
        headerToCopyFromName = listOfBiasNames[0]
    elif not headerToCopyFromName and not listOfBiasNames:
        raise Exception("Filename to copy header from not provided")

    masterBiasData = getMedianOfMatrices(listOfBiasData)
    # headerToCopyFromName is the file whose header we're copying to
    #  save in masterBias
    createFitFileWithSameHeader(masterBiasData, saveAs, headerToCopyFromName)

    return masterBiasData


#  makeMasterDark
#
#  purpose: creates masterDark, saves to fileName + returns masterDarkData
#
# we generate the masterDark by "taking median of the dark frames"
#   --Richard Berry, James Burnell
def makeMasterDark(
    saveAs=None, headerToCopyFromName=None, listOfDarkNames=None, listOfDarkData=None
):
    if listOfDarkNames:
        listOfDarkData = fitDataFromFitImages(listOfDarkNames)

    if not listOfDarkNames and not listOfDarkData:
        raise Exception("Neither Dark data nor names were provided")

    masterDarkData = getMedianOfMatrices(listOfDarkData)
    # listOfDarks[0] is the file whose header we're copying to
    #  save in masterDark

    # Create the file only if filename to save as is provided
    if saveAs is not None:
        if not headerToCopyFromName and listOfDarkNames:
            headerToCopyFromName = listOfDarkNames[0]
        elif not headerToCopyFromName and not listOfDarkNames:
            raise Exception("Filename to copy header from not provided")

        createFitFileWithSameHeader(masterDarkData, saveAs, headerToCopyFromName)

    return masterDarkData


#  makeMasterFlat
#
#  purpose: creates masterFlat, saves to fileName + returns masterFlatData
#
# we generate the masterFlat by
#   taking the median of flats and subtracting the masterDarkData
#
def makeMasterFlat(
    saveAs,
    masterDarkData,
    headerToCopyFromName=None,
    listOfFlatNames=None,
    listOfFlatData=None,
):
    # We're supposed to use flat dark for the master flat
    # but we did not take any for the new camera, so we're
    # using dark frames instead
    # In other words: If we don't have flat dark, use dark frames
    #

    if listOfFlatNames:
        listOfFlatNames = sorted_by_number(listOfFlatNames)
        listOfFlatData = fitDataFromFitImages(listOfFlatNames)

    if not listOfFlatNames and not listOfFlatData:
        raise Exception("Neither Flat data nor names were provided")

    if not headerToCopyFromName and listOfFlatNames:
        headerToCopyFromName = listOfFlatNames[0]
    elif not headerToCopyFromName and not listOfFlatNames:
        raise Exception("Filename to copy header from not provied")

    firstFlatMedian = customMedian(listOfFlatData[0])
    # We scale all flats w.r.t. first flat image
    #   like the current IDL code does
    #   https://github.com/LutherAstrophysics/idl-files/blob/f3d10e770d4d268908438deb4cda2076f21f1b14/master_calibration_frame_makerNEWEST.pro#L199
    listOfFlatData = [
        flatData * firstFlatMedian / customMedian(flatData) for flatData in listOfFlatData
    ]
    # the we take the median of the scaled flats
    combinedFlats = getMedianOfMatrices(listOfFlatData)
    masterFlatData = combinedFlats - masterDarkData

    # convert flat data to matrix of ints
    masterFlatData = np.array(masterFlatData, dtype="int")
    # listOfFlats[0] is the file whose header we're copying to
    #  save in masterDark
    createFitFileWithSameHeader(masterFlatData, saveAs, headerToCopyFromName)
    return masterFlatData


def getMedianOfMatrices(listOfMatrices):
    # https://stackoverflow.com/questions/18461623/average-values-in-two-numpy-arrays
    return customMedian(np.array(listOfMatrices), axis=0, out=np.empty_like(listOfMatrices[0]))



from datetime import date
from pathlib import Path

import numpy as np
import numpy.typing as npt


class NormfactorFile:
    # Class attributes
    date_format = "%m-%d-%y"

    @classmethod
    def generate_file_name(cls, night_date: date, img_duration: float):
        return f"{night_date.strftime(cls.date_format)}_m23_{img_duration}_normfactors.txt"

    def __init__(self, file_path: str) -> None:
        self.__path = Path(file_path)
        self.__data = None
        self.__is_read = False

    def _read(self):
        self.is_valid_file_name()
        if not self.exists():
            raise FileNotFoundError(f"File not found {self.__path}")
        with self.__path.open() as fd:
            lines = [line.strip() for line in fd.readlines()]
            self.__data = np.array(lines, dtype="float")
            self.__is_read = True

    def path(self):
        return self.__path

    def is_valid_file_name(self):
        if self.__path.is_dir() or self.__path.suffix != ".txt":
            return False
        return True

    def exists(self):
        return self.__path.exists()

    def data(self):
        if not self.__is_read:
            self._read()
        return self.__data

    def create_file(self, data: npt.NDArray):
        if not self.is_valid_file_name():
            raise ValueError(f"Invalid normfactor filename {self.path()}")
        with self.path().open("w") as fd:
            np.savetxt(
                fd,
                np.array(data),
                fmt="%3.5f",
            )

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"Normfactors {self.path()}"



import os


def rename(folder, newNameFromOldName, dry=True):
    """
    @param folder: folder where the the files to rename live
    @param: newNameFromOldName: function that takes old file name
            and returns new filename. You must pass this function.
    @param dry: specifies if you want to actually make the changes
    """
    files = os.listdir(folder)
    for file in files:
        if dry:
            print(f"{file} -> {newNameFromOldName(file)}")
        else:
            renameFile(file, folder, newNameFromOldName)


def renameFile(fileName, folderName, renameFunction):
    newName = renameFunction(fileName)
    oldFilePath = os.path.join(folderName, fileName)
    newFilePath = os.path.join(folderName, newName)
    os.rename(oldFilePath, newFilePath)
